"""TypeScript/JavaScript parser for Jac Lang."""

# ruff: noqa: ANN001, ANN201, ANN202, ANN204

from __future__ import annotations

import os
import sys
from collections.abc import Callable
from dataclasses import dataclass
from threading import Event
from typing import TYPE_CHECKING, Protocol, TypeAlias, TypeVar, cast

import jaclang.pycore.lark_ts_parser as ts_lark
import jaclang.pycore.unitree as uni
from jaclang.pycore.constant import TsTokens as Tok
from jaclang.pycore.passes import BaseTransform, Transform
from jaclang.vendor.lark.lexer import Token as LarkToken

if TYPE_CHECKING:
    from jaclang.pycore.program import JacProgram

T = TypeVar("T", bound=uni.UniNode)
TL = TypeVar("TL", bound=(uni.UniNode | list))


class _InteractiveParser(Protocol):
    def accepts(self) -> set[str]: ...

    def feed_token(self, token: object) -> None: ...


@dataclass
class TsLarkParseInput:
    """Input for TypeScript Lark parser transform."""

    ir_value: str
    on_error: Callable[[object], bool]


@dataclass
class TsLarkParseOutput:
    """Output from TypeScript Lark parser transform."""

    tree: object  # Lark Tree
    comments: list[object]  # Lark Tokens


class TsLarkParseTransform(BaseTransform[TsLarkParseInput, TsLarkParseOutput]):
    """Transform for TypeScript Lark parsing step."""

    comment_cache: list[object] = []

    def __init__(self, ir_in: TsLarkParseInput, prog: JacProgram) -> None:
        """Initialize TypeScript Lark parser transform."""
        super().__init__(ir_in=ir_in, prog=prog)

    def transform(self, ir_in: TsLarkParseInput) -> TsLarkParseOutput:
        """Transform input IR by parsing with LALR parser."""
        TsLarkParseTransform.comment_cache = []
        parser = ts_lark.Lark_StandAlone()
        try:
            tree = parser.parse(ir_in.ir_value, on_error=ir_in.on_error)
        except Exception as e:
            # Call error handler if provided
            if ir_in.on_error is not None:
                ir_in.on_error(e)
            raise
        return TsLarkParseOutput(
            tree=tree,
            comments=TsLarkParseTransform.comment_cache.copy(),
        )


class TypeScriptParser(Transform[uni.Source, uni.Module]):
    """TypeScript/JavaScript Parser."""

    def __init__(
        self,
        root_ir: uni.Source,
        prog: JacProgram,
        cancel_token: Event | None = None,
    ) -> None:
        """Initialize parser."""
        self.mod_path = root_ir.loc.mod_path
        self.node_list: list[uni.UniNode] = []
        self._node_ids: set[int] = set()

        if cancel_token and cancel_token.is_set():
            return
        Transform.__init__(self, ir_in=root_ir, prog=prog, cancel_token=cancel_token)

    def transform(self, ir_in: uni.Source) -> uni.Module:
        """Transform input IR."""
        try:
            lark_input = TsLarkParseInput(
                ir_value=ir_in.value,
                on_error=self.error_callback,
            )
            lark_transform = TsLarkParseTransform(ir_in=lark_input, prog=self.prog)
            parse_output = lark_transform.ir_out
            mod = TypeScriptParser.TreeToAST(parser=self).transform(parse_output.tree)
            ir_in.comments = [self.proc_comment(i, mod) for i in parse_output.comments]
            if not isinstance(mod, uni.Module):
                raise self.ice()
            if len(self.errors_had) != 0:
                mod.has_syntax_errors = True
                self.report_errors()
            self.ir_out = mod
            return mod
        except Exception as e:
            if hasattr(e, "line"):
                catch_error = self.error_to_token(e)
                error_msg = self.error_to_message(e)
                self.log_error(error_msg, node_override=catch_error)
            else:
                raise e

        self.report_errors()
        mod = uni.Module.make_stub(inject_src=ir_in)
        mod.has_syntax_errors = True
        return mod

    @staticmethod
    def proc_comment(token: object, mod: uni.UniNode) -> uni.CommentToken:
        """Process comment token."""
        return uni.CommentToken(
            orig_src=mod.loc.orig_src,
            name=getattr(token, "type", "COMMENT"),
            value=getattr(token, "value", str(token)),
            line=getattr(token, "line", 0) or 0,
            end_line=getattr(token, "end_line", 0) or 0,
            col_start=getattr(token, "column", 0) or 0,
            col_end=getattr(token, "end_column", 0) or 0,
            pos_start=getattr(token, "start_pos", 0) or 0,
            pos_end=getattr(token, "end_pos", 0) or 0,
            kid=[],
        )

    _MISSING_TOKENS = [
        Tok.SEMI,
        Tok.COMMA,
        Tok.COLON,
        Tok.RPAREN,
        Tok.RBRACE,
        Tok.RSQUARE,
    ]

    def error_callback(self, e: object) -> bool:
        """Handle parse error with recovery."""
        iparser = getattr(e, "interactive_parser", None)
        if iparser is None:
            return False
        iparser = cast(_InteractiveParser, iparser)

        def try_feed_missing_token(iparser: _InteractiveParser) -> Tok | None:
            """Try to feed a missing token to recover."""
            accepts = iparser.accepts()
            for tok in TypeScriptParser._MISSING_TOKENS:
                if tok.name in accepts:
                    iparser.feed_token(LarkToken(tok.name, tok.value))
                    return tok
            return None

        def feed_current_token(iparser: _InteractiveParser, tok: LarkToken) -> bool:
            """Feed current token after recovery."""
            max_attempts = 100
            attempts = 0
            while tok.type not in iparser.accepts():
                if attempts >= max_attempts:
                    return False
                if not try_feed_missing_token(iparser):
                    return False
                attempts += 1
            iparser.feed_token(tok)
            return True

        if hasattr(e, "token"):
            e_token = cast(LarkToken, e.token)
            if tk := try_feed_missing_token(iparser):
                self.log_error(f"Missing {tk.name}", self.error_to_token(e))
                return feed_current_token(iparser, e_token)

            self.log_error(
                f"Unexpected token '{e_token}'",
                self.error_to_token(e),
            )
            return True

        return False

    def error_to_message(self, e: object) -> str:
        """Convert error to message."""
        if hasattr(e, "token"):
            return f"Unexpected token '{cast(LarkToken, e.token)}'"
        return "Syntax Error"

    def error_to_token(self, e: object) -> uni.Token:
        """Convert error to token for location info."""
        catch_error = uni.EmptyToken()
        catch_error.orig_src = self.ir_in
        catch_error.line_no = getattr(e, "line", 1)
        catch_error.end_line = getattr(e, "line", 1)
        catch_error.c_start = getattr(e, "column", 0)
        catch_error.pos_start = getattr(e, "pos_in_stream", 0) or 0
        if hasattr(e, "token") and e.token:
            catch_error.c_end = getattr(e.token, "end_column", catch_error.c_start + 1)
            catch_error.pos_end = getattr(e.token, "end_pos", catch_error.pos_start + 1)
        else:
            catch_error.c_end = catch_error.c_start + 1
            catch_error.pos_end = catch_error.pos_start + 1
        return catch_error

    def report_errors(self, *, colors: bool = True) -> None:
        """Report errors to user."""
        from jaclang.pycore.helpers import ANSIColors

        if not sys.stderr.isatty():
            colors = False
        for alrt in self.errors_had:
            error_label = (
                "Error:" if not colors else f"{ANSIColors.RED}Error:{ANSIColors.END}"
            )
            print(error_label, end=" ", file=sys.stderr)
            print(alrt.pretty_print(colors=colors), file=sys.stderr)

    TsTransformer: TypeAlias = object  # Will be Lark Transformer

    class TreeToAST:
        """Transform TypeScript parse tree to AST."""

        def __init__(self, parser: TypeScriptParser) -> None:
            """Initialize transformer."""
            self.parse_ref = parser
            self.terminals: list[uni.Token] = []
            self.node_idx = 0
            self.cur_nodes: list[uni.UniNode] = []

        def transform(self, tree: object) -> uni.Module:
            """Transform the parse tree to AST."""
            lark_module = ts_lark
            transformer_base = lark_module.Transformer
            outer_ref = self

            # Dynamic class inheritance from Lark Transformer requires type: ignore
            # since Python's type system doesn't support runtime base classes
            class _InnerTransformer(transformer_base):  # type: ignore[misc, valid-type]
                """Inner Lark transformer."""

                def __init__(self) -> None:
                    super().__init__()
                    self.outer = outer_ref

                def _get_token(self, token: LarkToken) -> uni.Token:
                    """Convert Lark token to uni.Token."""
                    tok_type = token.type
                    tok_value = str(token)
                    tok_line = token.line or 1
                    tok_end_line = token.end_line or tok_line
                    tok_column = token.column or 0
                    tok_end_column = token.end_column or (tok_column + len(tok_value))
                    tok_start_pos = token.start_pos or 0
                    tok_end_pos = token.end_pos or (tok_start_pos + len(tok_value))
                    return uni.Token(
                        orig_src=self.outer.parse_ref.ir_in,
                        name=tok_type,
                        value=tok_value,
                        line=tok_line,
                        end_line=tok_end_line,
                        col_start=tok_column,
                        col_end=tok_end_column,
                        pos_start=tok_start_pos,
                        pos_end=tok_end_pos,
                    )

                def _make_name(self, token: LarkToken) -> uni.Name:
                    """Create a Name node from token."""
                    return uni.Name(
                        orig_src=self.outer.parse_ref.ir_in,
                        name=Tok.NAME.name,
                        value=str(token),
                        line=token.line or 1,
                        end_line=token.end_line or (token.line or 1),
                        col_start=token.column or 0,
                        col_end=token.end_column or (token.column or 0),
                        pos_start=token.start_pos or 0,
                        pos_end=token.end_pos or (token.start_pos or 0),
                    )

                def __default_token__(self, token: LarkToken) -> uni.Token:
                    """Default token handler."""
                    return self._get_token(token)

                def start(self, items: list[object]) -> uni.Module:
                    """start: source_file"""
                    if items and isinstance(items[0], uni.Module):
                        items[0]._in_mod_nodes = self.outer.parse_ref.node_list
                        return items[0]
                    return self.outer._make_module([])

                def source_file(self, items):
                    """source_file: statement*"""
                    stmts = [i for i in items if isinstance(i, uni.UniNode)]
                    return self.outer._make_module(stmts)

                def statement(self, items):
                    """statement: declaration | ..."""
                    return items[0] if items else None

                def declaration(self, items):
                    """declaration: var_decl | func_decl | ..."""
                    return items[0] if items else None

                # Variable Declarations
                def var_decl(self, items):
                    """var_decl: (KW_VAR | KW_LET | KW_CONST) binding_list semi"""
                    kids = [i for i in items if i is not None]
                    keyword = kids[0] if kids else None
                    is_frozen = False
                    if isinstance(keyword, uni.Token):
                        is_frozen = keyword.name in (Tok.KW_CONST.name, "KW_CONST")

                    bindings = []
                    for item in kids[1:]:
                        if isinstance(item, list):
                            bindings.extend(item)
                        elif isinstance(item, uni.Assignment):
                            bindings.append(item)

                    return uni.GlobalVars(
                        access=None,
                        assignments=bindings,
                        is_frozen=is_frozen,
                        kid=kids,
                    )

                def binding_list(self, items):
                    """binding_list: binding (COMMA binding)*"""
                    return [i for i in items if isinstance(i, uni.Assignment)]

                def binding(self, items):
                    """binding: binding_pattern type_annotation? initializer?"""
                    kids = [i for i in items if i is not None]
                    name = None
                    type_tag = None
                    value = None

                    for item in kids:
                        if isinstance(item, (uni.Name, uni.Token)) and name is None:
                            if (
                                isinstance(item, uni.Token)
                                and item.name == Tok.NAME.name
                            ):
                                name = uni.Name(
                                    orig_src=item.orig_src,
                                    name=item.name,
                                    value=item.value,
                                    line=item.line_no,
                                    end_line=item.end_line,
                                    col_start=item.c_start,
                                    col_end=item.c_end,
                                    pos_start=item.pos_start,
                                    pos_end=item.pos_end,
                                )
                            elif isinstance(item, uni.Name):
                                name = item
                        elif isinstance(item, uni.SubTag):
                            type_tag = item
                        elif isinstance(item, uni.Expr):
                            value = item

                    if name is None:
                        return None

                    return uni.Assignment(
                        target=uni.AtomTrailer(
                            target=name,
                            right=None,
                            is_null_ok=False,
                            is_attr=False,
                            kid=[name],
                        ),
                        type_tag=type_tag,
                        value=value,
                        kid=kids,
                    )

                def binding_pattern(self, items):
                    """binding_pattern: NAME | array_binding | object_binding"""
                    if items:
                        item = items[0]
                        if hasattr(item, "value"):
                            return self._make_name(item)
                        return item
                    return None

                def type_annotation(self, items):
                    """type_annotation: COLON type"""
                    kids = [i for i in items if i is not None]
                    if len(kids) >= 2:
                        type_expr = kids[1] if isinstance(kids[1], uni.Expr) else None
                        if type_expr:
                            return uni.SubTag[uni.Expr](tag=type_expr, kid=kids)
                    return None

                def initializer(self, items):
                    """initializer: EQ assignment_expr"""
                    if len(items) >= 2:
                        return items[1]
                    return items[0] if items else None

                # Function Declarations
                def func_decl(self, items):
                    """func_decl: KW_ASYNC? KW_FUNCTION STAR? NAME? ..."""
                    kids = [i for i in items if i is not None]
                    is_async = False
                    name = None
                    params = []
                    return_type = None
                    body = []

                    for item in kids:
                        if isinstance(item, uni.Token):
                            if item.name in (Tok.KW_ASYNC.name, "KW_ASYNC"):
                                is_async = True
                            elif item.name in (Tok.NAME.name, "NAME") and name is None:
                                name = self._make_name(item)
                        elif isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, list):
                            if not params:
                                params = [
                                    p for p in item if isinstance(p, uni.ParamVar)
                                ]
                            else:
                                body = [
                                    s for s in item if isinstance(s, uni.CodeBlockStmt)
                                ]
                        elif isinstance(item, uni.SubTag):
                            return_type = item

                    if name is None:
                        name = uni.Name.gen_stub_from_node(
                            self.outer.parse_ref.ir_in, "[anonymous]"
                        )

                    sig = uni.FuncSignature(
                        params=params,
                        return_type=return_type.tag if return_type else None,
                        kid=kids,
                    )

                    return uni.Ability(
                        name_ref=name,
                        is_func=True,
                        is_async=is_async,
                        is_override=False,
                        is_static=False,
                        is_abstract=False,
                        access=None,
                        signature=sig,
                        body=body,
                        decorators=None,
                        doc=None,
                        kid=kids,
                    )

                def param_list(self, items):
                    """param_list: param (COMMA param)*"""
                    return [i for i in items if isinstance(i, uni.ParamVar)]

                def param(self, items):
                    """param: binding_pattern QUESTION? type_annotation? initializer?"""
                    kids = [i for i in items if i is not None]
                    name = None
                    type_tag = None
                    default_val = None

                    for item in kids:
                        if isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, uni.Token):
                            if item.name in (Tok.NAME.name, "NAME") and name is None:
                                name = self._make_name(item)
                            # QUESTION token indicates optional param (currently unused)
                        elif isinstance(item, uni.SubTag):
                            type_tag = item
                        elif isinstance(item, uni.Expr) and default_val is None:
                            default_val = item

                    if name is None:
                        return None

                    return uni.ParamVar(
                        name=name,
                        unpack=None,
                        type_tag=type_tag,
                        value=default_val,
                        kid=kids,
                    )

                def func_body(self, items):
                    """func_body: block_stmt"""
                    return items[0] if items else []

                # Class Declarations
                def class_decl(self, items):
                    """class_decl: decorators? KW_ABSTRACT? KW_CLASS NAME? ..."""
                    kids = [i for i in items if i is not None]
                    name = None
                    base_classes = []
                    body = []
                    decorators = []

                    for item in kids:
                        if isinstance(item, uni.Token):
                            # KW_ABSTRACT token indicates abstract class (currently unused)
                            if item.name in (Tok.NAME.name, "NAME") and name is None:
                                name = self._make_name(item)
                        elif isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, list):
                            if all(isinstance(i, uni.Expr) for i in item):
                                decorators = item
                            elif all(isinstance(i, uni.ArchBlockStmt) for i in item):
                                body = item
                            else:
                                body = [
                                    i for i in item if isinstance(i, uni.ArchBlockStmt)
                                ]

                    if name is None:
                        name = uni.Name.gen_stub_from_node(
                            self.outer.parse_ref.ir_in, "[anonymous]"
                        )

                    return uni.Archetype(
                        name=name,
                        arch_type=uni.Token(
                            orig_src=self.outer.parse_ref.ir_in,
                            name="KW_CLASS",
                            value="class",
                            line=1,
                            end_line=1,
                            col_start=0,
                            col_end=5,
                            pos_start=0,
                            pos_end=5,
                        ),
                        access=None,
                        base_classes=base_classes,
                        body=body,
                        decorators=decorators if decorators else None,
                        doc=None,
                        kid=kids,
                    )

                def class_body(self, items):
                    """class_body: LBRACE class_element* RBRACE"""
                    return [i for i in items if isinstance(i, uni.ArchBlockStmt)]

                def class_element(self, items):
                    """class_element: decorators? class_element_modifiers? class_member"""
                    if items:
                        for item in items:
                            if isinstance(item, uni.ArchBlockStmt):
                                return item
                    return None

                def class_member(self, items):
                    """class_member: property_definition | method_definition | ..."""
                    return items[0] if items else None

                def property_definition(self, items):
                    """property_definition: property_name ...? type_annotation? initializer?"""
                    kids = [i for i in items if i is not None]
                    name = None
                    type_tag = None
                    value = None

                    for item in kids:
                        if isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, uni.Token) and name is None:
                            if item.name in (Tok.NAME.name, "NAME"):
                                name = self._make_name(item)
                        elif isinstance(item, uni.SubTag):
                            type_tag = item
                        elif isinstance(item, uni.Expr):
                            value = item

                    if name is None:
                        return None

                    return uni.HasVar(
                        name=name,
                        type_tag=type_tag,
                        value=value,
                        is_static=False,
                        is_frozen=False,
                        access=None,
                        doc=None,
                        kid=kids,
                    )

                def method_definition(self, items):
                    """method_definition: KW_ASYNC? STAR? property_name ..."""
                    kids = [i for i in items if i is not None]
                    is_async = False
                    name = None
                    params = []
                    return_type = None
                    body = []

                    for item in kids:
                        if isinstance(item, uni.Token):
                            if item.name in (Tok.KW_ASYNC.name, "KW_ASYNC"):
                                is_async = True
                            # STAR token indicates generator (currently unused)
                            elif item.name in (Tok.NAME.name, "NAME") and name is None:
                                name = self._make_name(item)
                        elif isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, list):
                            if all(isinstance(p, uni.ParamVar) for p in item if p):
                                params = [
                                    p for p in item if isinstance(p, uni.ParamVar)
                                ]
                            else:
                                body = [
                                    s for s in item if isinstance(s, uni.CodeBlockStmt)
                                ]
                        elif isinstance(item, uni.SubTag):
                            return_type = item

                    if name is None:
                        return None

                    sig = uni.FuncSignature(
                        params=params,
                        return_type=return_type.tag if return_type else None,
                        kid=kids,
                    )

                    return uni.Ability(
                        name_ref=name,
                        is_func=True,
                        is_async=is_async,
                        is_override=False,
                        is_static=False,
                        is_abstract=False,
                        access=None,
                        signature=sig,
                        body=body,
                        decorators=None,
                        doc=None,
                        kid=kids,
                    )

                # Interface Declarations
                def interface_decl(self, items):
                    """interface_decl: KW_INTERFACE NAME type_params? ..."""
                    kids = [i for i in items if i is not None]
                    name = None
                    body = []

                    for item in kids:
                        if isinstance(item, uni.Token):
                            if item.name in (Tok.NAME.name, "NAME") and name is None:
                                name = self._make_name(item)
                        elif isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, list):
                            body = [i for i in item if isinstance(i, uni.ArchBlockStmt)]

                    if name is None:
                        name = uni.Name.gen_stub_from_node(
                            self.outer.parse_ref.ir_in, "[anonymous]"
                        )

                    return uni.Archetype(
                        name=name,
                        arch_type=uni.Token(
                            orig_src=self.outer.parse_ref.ir_in,
                            name="KW_INTERFACE",
                            value="interface",
                            line=1,
                            end_line=1,
                            col_start=0,
                            col_end=9,
                            pos_start=0,
                            pos_end=9,
                        ),
                        access=None,
                        base_classes=[],
                        body=body,
                        decorators=None,
                        doc=None,
                        kid=kids,
                    )

                # Type Alias
                def type_alias_decl(self, items):
                    """type_alias_decl: KW_TYPE NAME type_params? EQ type semi"""
                    kids = [i for i in items if i is not None]
                    name = None
                    type_val = None

                    for item in kids:
                        if isinstance(item, uni.Token):
                            if item.name in (Tok.NAME.name, "NAME") and name is None:
                                name = self._make_name(item)
                        elif isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, uni.Expr):
                            type_val = item

                    if name is None:
                        return None

                    # Use Assignment to represent type alias
                    return uni.Assignment(
                        target=uni.AtomTrailer(
                            target=name,
                            right=None,
                            is_null_ok=False,
                            is_attr=False,
                            kid=[name],
                        ),
                        type_tag=None,
                        value=type_val,
                        kid=kids,
                    )

                # Enum Declarations
                def enum_decl(self, items):
                    """enum_decl: KW_CONST? KW_ENUM NAME LBRACE enum_member_list? RBRACE"""
                    kids = [i for i in items if i is not None]
                    name = None
                    members = []

                    for item in kids:
                        if isinstance(item, uni.Token):
                            if item.name in (Tok.NAME.name, "NAME") and name is None:
                                name = self._make_name(item)
                        elif isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, list):
                            members = [m for m in item if isinstance(m, uni.Assignment)]

                    if name is None:
                        name = uni.Name.gen_stub_from_node(
                            self.outer.parse_ref.ir_in, "[anonymous]"
                        )

                    return uni.Enum(
                        name=name,
                        access=None,
                        base_classes=[],
                        body=members,
                        decorators=None,
                        doc=None,
                        kid=kids,
                    )

                def enum_member_list(self, items):
                    """enum_member_list: enum_member (COMMA enum_member)*"""
                    return [i for i in items if isinstance(i, uni.Assignment)]

                def enum_member(self, items):
                    """enum_member: property_name (EQ assignment_expr)?"""
                    kids = [i for i in items if i is not None]
                    name = None
                    value = None

                    for item in kids:
                        if isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, uni.Token):
                            if item.name in (Tok.NAME.name, "NAME") and name is None:
                                name = self._make_name(item)
                        elif isinstance(item, uni.Expr):
                            value = item

                    if name is None:
                        return None

                    return uni.Assignment(
                        target=uni.AtomTrailer(
                            target=name,
                            right=None,
                            is_null_ok=False,
                            is_attr=False,
                            kid=[name],
                        ),
                        type_tag=None,
                        value=value,
                        kid=kids,
                    )

                # Import/Export
                def import_decl(self, items):
                    """import_decl: KW_IMPORT ..."""
                    kids = [i for i in items if i is not None]
                    from_loc = None
                    import_items = []

                    for item in kids:
                        if isinstance(item, uni.Token):
                            pass  # KW_TYPE token indicates type-only import (currently unused)
                        elif isinstance(item, uni.ModulePath):
                            from_loc = item
                        elif isinstance(item, list):
                            import_items = [
                                i for i in item if isinstance(i, uni.ModuleItem)
                            ]
                        elif isinstance(item, uni.ModuleItem):
                            import_items.append(item)

                    return uni.Import(
                        from_loc=from_loc,
                        items=import_items
                        if import_items
                        else [from_loc]
                        if from_loc
                        else [],
                        is_absorb=False,
                        kid=kids,
                    )

                def from_clause(self, items):
                    """from_clause: KW_FROM module_specifier"""
                    for item in items:
                        if isinstance(item, uni.ModulePath):
                            return item
                        if isinstance(item, uni.Token) and item.name in (
                            "STRING",
                            Tok.STRING.name,
                        ):
                            path_val = item.value.strip("'\"")
                            return uni.ModulePath(
                                path=[
                                    uni.Name(
                                        orig_src=item.orig_src,
                                        name="NAME",
                                        value=path_val,
                                        line=item.line_no,
                                        end_line=item.end_line,
                                        col_start=item.c_start,
                                        col_end=item.c_end,
                                        pos_start=item.pos_start,
                                        pos_end=item.pos_end,
                                    )
                                ],
                                level=0,
                                alias=None,
                                kid=[item],
                            )
                    return None

                def module_specifier(self, items):
                    """module_specifier: STRING"""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Token):
                            path_val = item.value.strip("'\"")
                            return uni.ModulePath(
                                path=[
                                    uni.Name(
                                        orig_src=item.orig_src,
                                        name="NAME",
                                        value=path_val,
                                        line=item.line_no,
                                        end_line=item.end_line,
                                        col_start=item.c_start,
                                        col_end=item.c_end,
                                        pos_start=item.pos_start,
                                        pos_end=item.pos_end,
                                    )
                                ],
                                level=0,
                                alias=None,
                                kid=[item],
                            )
                    return None

                def named_imports(self, items):
                    """named_imports: LBRACE import_specifier_list? RBRACE"""
                    result = []
                    for item in items:
                        if isinstance(item, list):
                            result.extend(
                                [i for i in item if isinstance(i, uni.ModuleItem)]
                            )
                        elif isinstance(item, uni.ModuleItem):
                            result.append(item)
                    return result

                def import_specifier_list(self, items):
                    """import_specifier_list: import_specifier (COMMA import_specifier)*"""
                    return [i for i in items if isinstance(i, uni.ModuleItem)]

                def import_specifier(self, items):
                    """import_specifier: NAME (KW_AS NAME)?"""
                    kids = [i for i in items if i is not None]
                    name = None
                    alias = None

                    for item in kids:
                        if isinstance(item, uni.Token) and item.name in (
                            Tok.NAME.name,
                            "NAME",
                        ):
                            if name is None:
                                name = self._make_name(item)
                            else:
                                alias = self._make_name(item)

                    if name is None:
                        return None

                    return uni.ModuleItem(
                        name=name,
                        alias=alias,
                        kid=kids,
                    )

                def export_decl(self, items):
                    """export_decl: KW_EXPORT ..."""
                    # For simplicity, return the inner declaration
                    for item in items:
                        if isinstance(item, uni.UniNode) and not isinstance(
                            item, uni.Token
                        ):
                            return item
                    return None

                # Control Flow
                def block_stmt(self, items):
                    """block_stmt: LBRACE statement* RBRACE"""
                    return [i for i in items if isinstance(i, uni.CodeBlockStmt)]

                def if_stmt(self, items):
                    """if_stmt: KW_IF LPAREN expression RPAREN statement (KW_ELSE statement)?"""
                    kids = [i for i in items if i is not None]
                    condition = None
                    body = []
                    else_body = None

                    for item in kids:
                        if isinstance(item, uni.Expr) and condition is None:
                            condition = item
                        elif isinstance(item, uni.CodeBlockStmt):
                            if not body:
                                body = [item]
                            else:
                                else_body = uni.ElseStmt(body=[item], kid=[item])
                        elif isinstance(item, list):
                            if not body:
                                body = [
                                    i for i in item if isinstance(i, uni.CodeBlockStmt)
                                ]
                            else:
                                else_stmts = [
                                    i for i in item if isinstance(i, uni.CodeBlockStmt)
                                ]
                                if else_stmts:
                                    else_body = uni.ElseStmt(
                                        body=else_stmts, kid=else_stmts
                                    )

                    return uni.IfStmt(
                        condition=condition,
                        body=body,
                        else_body=else_body,
                        kid=kids,
                    )

                def while_stmt(self, items):
                    """while_stmt: KW_WHILE LPAREN expression RPAREN statement"""
                    kids = [i for i in items if i is not None]
                    condition = None
                    body = []

                    for item in kids:
                        if isinstance(item, uni.Expr) and condition is None:
                            condition = item
                        elif isinstance(item, uni.CodeBlockStmt):
                            body = [item]
                        elif isinstance(item, list):
                            body = [i for i in item if isinstance(i, uni.CodeBlockStmt)]

                    return uni.WhileStmt(
                        condition=condition,
                        body=body,
                        kid=kids,
                    )

                def for_stmt(self, items):
                    """for_stmt: KW_FOR LPAREN for_init? SEMI expression? SEMI expression? RPAREN statement"""
                    kids = [i for i in items if i is not None]
                    body = []

                    for item in kids:
                        if isinstance(item, list):
                            body = [i for i in item if isinstance(i, uni.CodeBlockStmt)]
                        elif isinstance(item, uni.CodeBlockStmt):
                            body = [item]

                    # Create an IterForStmt as approximation
                    return uni.IterForStmt(
                        iter=None,
                        condition=None,
                        count_by=None,
                        body=body,
                        else_body=None,
                        is_async=False,
                        kid=kids,
                    )

                def for_in_stmt(self, items):
                    """for_in_stmt: KW_FOR ..."""
                    return self.for_stmt(items)

                def for_of_stmt(self, items):
                    """for_of_stmt: KW_FOR ..."""
                    return self.for_stmt(items)

                def return_stmt(self, items):
                    """return_stmt: KW_RETURN expression? semi"""
                    kids = [i for i in items if i is not None]
                    expr = None
                    for item in kids:
                        if isinstance(item, uni.Expr):
                            expr = item
                            break

                    return uni.ReturnStmt(expr=expr, kid=kids)

                def try_stmt(self, items):
                    """try_stmt: KW_TRY block_stmt catch_clause? finally_clause?"""
                    kids = [i for i in items if i is not None]
                    body = []
                    excepts = []
                    finally_body = None

                    for item in kids:
                        if isinstance(item, list):
                            if not body:
                                body = [
                                    i for i in item if isinstance(i, uni.CodeBlockStmt)
                                ]
                        elif isinstance(item, uni.Except):
                            excepts.append(item)
                        elif isinstance(item, uni.FinallyStmt):
                            finally_body = item

                    return uni.TryStmt(
                        body=body,
                        excepts=excepts,
                        finally_body=finally_body,
                        kid=kids,
                    )

                def catch_clause(self, items):
                    """catch_clause: KW_CATCH (LPAREN binding_pattern type_annotation? RPAREN)? block_stmt"""
                    kids = [i for i in items if i is not None]
                    name = None
                    body = []

                    for item in kids:
                        if isinstance(item, uni.Name) and name is None:
                            name = item
                        elif isinstance(item, uni.Token) and name is None:
                            if item.name in (Tok.NAME.name, "NAME"):
                                name = self._make_name(item)
                        elif isinstance(item, list):
                            body = [i for i in item if isinstance(i, uni.CodeBlockStmt)]

                    return uni.Except(
                        ex_type=None,
                        name=name,
                        body=body,
                        kid=kids,
                    )

                def finally_clause(self, items):
                    """finally_clause: KW_FINALLY block_stmt"""
                    kids = [i for i in items if i is not None]
                    body = []
                    for item in kids:
                        if isinstance(item, list):
                            body = [i for i in item if isinstance(i, uni.CodeBlockStmt)]

                    return uni.FinallyStmt(body=body, kid=kids)

                def throw_stmt(self, items):
                    """throw_stmt: KW_THROW expression semi"""
                    kids = [i for i in items if i is not None]
                    expr = None
                    for item in kids:
                        if isinstance(item, uni.Expr):
                            expr = item
                            break

                    return uni.RaiseStmt(cause=expr, from_target=None, kid=kids)

                def break_stmt(self, items):
                    """break_stmt: KW_BREAK NAME? semi"""
                    return uni.CtrlStmt(
                        ctrl=uni.Token(
                            orig_src=self.outer.parse_ref.ir_in,
                            name="KW_BREAK",
                            value="break",
                            line=1,
                            end_line=1,
                            col_start=0,
                            col_end=5,
                            pos_start=0,
                            pos_end=5,
                        ),
                        kid=[i for i in items if i is not None],
                    )

                def continue_stmt(self, items):
                    """continue_stmt: KW_CONTINUE NAME? semi"""
                    return uni.CtrlStmt(
                        ctrl=uni.Token(
                            orig_src=self.outer.parse_ref.ir_in,
                            name="KW_CONTINUE",
                            value="continue",
                            line=1,
                            end_line=1,
                            col_start=0,
                            col_end=8,
                            pos_start=0,
                            pos_end=8,
                        ),
                        kid=[i for i in items if i is not None],
                    )

                def empty_stmt(self, items):
                    """empty_stmt: SEMI"""
                    return None

                def expr_stmt(self, items):
                    """expr_stmt: expression semi"""
                    kids = [i for i in items if i is not None]
                    for item in kids:
                        if isinstance(item, uni.Expr):
                            return uni.ExprStmt(expr=item, kid=kids)
                    return None

                # Expressions
                def expression(self, items):
                    """expression: assignment_expr (COMMA assignment_expr)*"""
                    exprs = [i for i in items if isinstance(i, uni.Expr)]
                    if len(exprs) == 1:
                        return exprs[0]
                    elif len(exprs) > 1:
                        return uni.TupleVal(values=exprs, kid=list(items))
                    return None

                def assignment_expr(self, items):
                    """assignment_expr: conditional_expr | left_hand_side_expr assignment_op assignment_expr"""
                    kids = [i for i in items if i is not None]
                    if len(kids) == 1:
                        return kids[0] if isinstance(kids[0], uni.Expr) else None

                    # Assignment case
                    if len(kids) >= 3:
                        target = kids[0]
                        # kids[1] is the assignment operator (unused)
                        value = kids[2]
                        if isinstance(target, uni.Expr) and isinstance(value, uni.Expr):
                            return uni.Assignment(
                                target=target,
                                type_tag=None,
                                value=value,
                                kid=kids,
                            )
                    return kids[0] if kids and isinstance(kids[0], uni.Expr) else None

                def conditional_expr(self, items):
                    """conditional_expr: nullish_coalescing_expr (QUESTION assignment_expr COLON assignment_expr)?"""
                    kids = [i for i in items if i is not None]
                    exprs = [i for i in kids if isinstance(i, uni.Expr)]

                    if len(exprs) == 1:
                        return exprs[0]
                    elif len(exprs) == 3:
                        return uni.IfElseExpr(
                            condition=exprs[0],
                            value=exprs[1],
                            else_value=exprs[2],
                            kid=kids,
                        )
                    return exprs[0] if exprs else None

                def _binary_expr(self, items, op_name: str = "OP"):
                    """Helper for binary expressions."""
                    kids = [i for i in items if i is not None]
                    exprs = [i for i in kids if isinstance(i, uni.Expr)]
                    ops = [i for i in kids if isinstance(i, uni.Token)]

                    if len(exprs) == 1:
                        return exprs[0]
                    elif len(exprs) >= 2 and ops:
                        return uni.BinaryExpr(
                            left=exprs[0],
                            op=ops[0],
                            right=exprs[1],
                            kid=kids,
                        )
                    return exprs[0] if exprs else None

                def nullish_coalescing_expr(self, items):
                    return self._binary_expr(items, "NULLISH_COALESCE")

                def logical_or_expr(self, items):
                    return self._binary_expr(items, "OR")

                def logical_and_expr(self, items):
                    return self._binary_expr(items, "AND")

                def bitwise_or_expr(self, items):
                    return self._binary_expr(items, "BW_OR")

                def bitwise_xor_expr(self, items):
                    return self._binary_expr(items, "BW_XOR")

                def bitwise_and_expr(self, items):
                    return self._binary_expr(items, "BW_AND")

                def equality_expr(self, items):
                    return self._binary_expr(items, "EQ")

                def relational_expr(self, items):
                    return self._binary_expr(items, "CMP")

                def shift_expr(self, items):
                    return self._binary_expr(items, "SHIFT")

                def additive_expr(self, items):
                    return self._binary_expr(items, "ADD")

                def multiplicative_expr(self, items):
                    return self._binary_expr(items, "MUL")

                def exponentiation_expr(self, items):
                    return self._binary_expr(items, "EXP")

                def unary_expr(self, items):
                    """unary_expr: postfix_expr | (op) unary_expr"""
                    kids = [i for i in items if i is not None]
                    if len(kids) == 1:
                        return kids[0] if isinstance(kids[0], uni.Expr) else None

                    # Unary operation
                    op = None
                    operand = None
                    for item in kids:
                        if isinstance(item, uni.Token) and op is None:
                            op = item
                        elif isinstance(item, uni.Expr):
                            operand = item

                    if op and operand:
                        return uni.UnaryExpr(op=op, operand=operand, kid=kids)
                    return operand

                def postfix_expr(self, items):
                    """postfix_expr: left_hand_side_expr (PLUS_PLUS | MINUS_MINUS)?"""
                    kids = [i for i in items if i is not None]
                    expr = None
                    op = None
                    for item in kids:
                        if isinstance(item, uni.Expr) and expr is None:
                            expr = item
                        elif isinstance(item, uni.Token):
                            op = item

                    if op and expr:
                        return uni.UnaryExpr(op=op, operand=expr, kid=kids)
                    return expr

                def left_hand_side_expr(self, items):
                    """left_hand_side_expr: call_expr | new_expr"""
                    return items[0] if items else None

                def new_expr(self, items):
                    """new_expr: member_expr | KW_NEW new_expr"""
                    return items[-1] if items else None

                def call_expr(self, items):
                    """call_expr: member_expr arguments | ..."""
                    kids = [i for i in items if i is not None]
                    target = None
                    args = []

                    for item in kids:
                        if isinstance(item, uni.Expr) and target is None:
                            target = item
                        elif isinstance(item, list):
                            args = [a for a in item if isinstance(a, uni.Expr)]

                    if target:
                        if args or any(isinstance(k, list) for k in kids):
                            return uni.FuncCall(
                                target=target,
                                params=args,
                                kid=kids,
                            )
                        return target
                    return None

                def member_expr(self, items):
                    """member_expr: primary_expr | member_expr DOT NAME | member_expr LSQUARE expression RSQUARE"""
                    kids = [i for i in items if i is not None]
                    if len(kids) == 1:
                        return kids[0] if isinstance(kids[0], uni.Expr) else None

                    # Member access
                    target = None
                    accessor = None
                    for item in kids:
                        if isinstance(item, uni.Expr) and target is None:
                            target = item
                        elif isinstance(item, (uni.Name, uni.Token)):
                            if isinstance(item, uni.Token) and item.name in (
                                Tok.NAME.name,
                                "NAME",
                            ):
                                accessor = self._make_name(item)
                            elif isinstance(item, uni.Name):
                                accessor = item
                        elif isinstance(item, uni.Expr) and target is not None:
                            accessor = item

                    if target and accessor:
                        return uni.AtomTrailer(
                            target=target,
                            right=accessor,
                            is_null_ok=False,
                            is_attr=isinstance(accessor, uni.Name),
                            kid=kids,
                        )
                    return target

                def arguments(self, items):
                    """arguments: type_args? LPAREN argument_list? RPAREN"""
                    for item in items:
                        if isinstance(item, list):
                            return item
                    return []

                def argument_list(self, items):
                    """argument_list: argument (COMMA argument)*"""
                    return [i for i in items if isinstance(i, uni.Expr)]

                def argument(self, items):
                    """argument: ELLIPSIS? assignment_expr"""
                    for item in items:
                        if isinstance(item, uni.Expr):
                            return item
                    return None

                def primary_expr(self, items):
                    """primary_expr: KW_THIS | NAME | literal | ..."""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Token):
                            if item.name in (Tok.NAME.name, "NAME"):
                                return self._make_name(item)
                            elif item.name in (
                                Tok.KW_THIS.name,
                                "KW_THIS",
                            ) or item.name in (Tok.KW_SUPER.name, "KW_SUPER"):
                                return uni.SpecialVarRef(
                                    var=item,
                                    kid=[item],
                                )
                        return item if isinstance(item, uni.Expr) else None
                    return None

                def paren_expr(self, items):
                    """paren_expr: LPAREN expression RPAREN"""
                    for item in items:
                        if isinstance(item, uni.Expr):
                            return item
                    return None

                # Literals
                def literal(self, items):
                    """literal: NULL | TRUE | FALSE | NUMBER | STRING | REGEX"""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Token):
                            if item.name in ("NULL", Tok.NULL.name):
                                return uni.Null(kid=[item])
                            elif item.name in (
                                "TRUE",
                                "FALSE",
                                Tok.TRUE.name,
                                Tok.FALSE.name,
                            ):
                                return uni.Bool(
                                    value=item.value.lower() == "true",
                                    kid=[item],
                                )
                            elif item.name in ("NUMBER", Tok.NUMBER.name):
                                if "." in item.value or "e" in item.value.lower():
                                    return uni.Float(
                                        value=item.value,
                                        kid=[item],
                                    )
                                else:
                                    return uni.Int(
                                        value=item.value,
                                        kid=[item],
                                    )
                            elif item.name in (
                                "STRING",
                                Tok.STRING.name,
                            ) or item.name in ("REGEX", Tok.REGEX.name):
                                return uni.String(
                                    value=item.value,
                                    kid=[item],
                                )
                        return item if isinstance(item, uni.Expr) else None
                    return None

                def array_literal(self, items):
                    """array_literal: LSQUARE element_list? RSQUARE"""
                    elements = []
                    for item in items:
                        if isinstance(item, list):
                            elements = [e for e in item if isinstance(e, uni.Expr)]
                        elif isinstance(item, uni.Expr):
                            elements.append(item)
                    return uni.ListVal(values=elements, kid=list(items))

                def element_list(self, items):
                    """element_list: array_element (COMMA array_element)*"""
                    return [i for i in items if isinstance(i, uni.Expr)]

                def array_element(self, items):
                    """array_element: ELLIPSIS? assignment_expr | COMMA"""
                    for item in items:
                        if isinstance(item, uni.Expr):
                            return item
                    return None

                def object_literal(self, items):
                    """object_literal: LBRACE property_list? RBRACE"""
                    props = []
                    for item in items:
                        if isinstance(item, list):
                            props = [p for p in item if isinstance(p, uni.KVPair)]
                    return uni.DictVal(kv_pairs=props, kid=list(items))

                def property_list(self, items):
                    """property_list: property_definition_expr (COMMA property_definition_expr)*"""
                    return [i for i in items if isinstance(i, uni.KVPair)]

                def property_definition_expr(self, items):
                    """property_definition_expr: property_name COLON assignment_expr | ..."""
                    kids = [i for i in items if i is not None]
                    key = None
                    value = None

                    for item in kids:
                        if isinstance(item, uni.Name) and key is None:
                            key = item
                        elif isinstance(item, uni.Token) and key is None:
                            if item.name in (Tok.NAME.name, "NAME"):
                                key = self._make_name(item)
                        elif isinstance(item, uni.Expr):
                            if key is None:
                                key = item
                            else:
                                value = item

                    if key:
                        if value is None:
                            value = key
                        return uni.KVPair(key=key, value=value, kid=kids)
                    return None

                def property_name(self, items):
                    """property_name: NAME | STRING | NUMBER | computed_property_name"""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Token):
                            if item.name in (Tok.NAME.name, "NAME"):
                                return self._make_name(item)
                            elif item.name in (Tok.STRING.name, "STRING"):
                                return uni.String(value=item.value, kid=[item])
                            elif item.name in (Tok.NUMBER.name, "NUMBER"):
                                return uni.Int(value=item.value, kid=[item])
                        return item if isinstance(item, uni.Expr) else None
                    return None

                # Arrow Functions
                def arrow_func(self, items):
                    """arrow_func: KW_ASYNC? type_params? arrow_params type_annotation? ARROW arrow_body"""
                    kids = [i for i in items if i is not None]
                    is_async = False
                    params = []
                    return_type = None
                    body = []

                    for item in kids:
                        if isinstance(item, uni.Token):
                            if item.name in (Tok.KW_ASYNC.name, "KW_ASYNC"):
                                is_async = True
                            elif item.name in (Tok.NAME.name, "NAME"):
                                # Single parameter shorthand
                                params = [
                                    uni.ParamVar(
                                        name=self._make_name(item),
                                        unpack=None,
                                        type_tag=None,
                                        value=None,
                                        kid=[item],
                                    )
                                ]
                        elif isinstance(item, list):
                            if all(isinstance(p, uni.ParamVar) for p in item if p):
                                params = [
                                    p for p in item if isinstance(p, uni.ParamVar)
                                ]
                            else:
                                body = [
                                    s for s in item if isinstance(s, uni.CodeBlockStmt)
                                ]
                        elif isinstance(item, uni.SubTag):
                            return_type = item
                        elif isinstance(item, uni.Expr):
                            # Expression body - wrap in implicit return
                            body = [uni.ReturnStmt(expr=item, kid=[item])]

                    name = uni.Name.gen_stub_from_node(
                        self.outer.parse_ref.ir_in, "[arrow]"
                    )

                    sig = uni.FuncSignature(
                        params=params,
                        return_type=return_type.tag if return_type else None,
                        kid=kids,
                    )

                    return uni.Ability(
                        name_ref=name,
                        is_func=True,
                        is_async=is_async,
                        is_override=False,
                        is_static=False,
                        is_abstract=False,
                        access=None,
                        signature=sig,
                        body=body,
                        decorators=None,
                        doc=None,
                        kid=kids,
                    )

                def arrow_params(self, items):
                    """arrow_params: NAME | LPAREN param_list? RPAREN"""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Token) and item.name in (
                            Tok.NAME.name,
                            "NAME",
                        ):
                            return [
                                uni.ParamVar(
                                    name=self._make_name(item),
                                    unpack=None,
                                    type_tag=None,
                                    value=None,
                                    kid=[item],
                                )
                            ]
                        elif isinstance(item, list):
                            return [p for p in item if isinstance(p, uni.ParamVar)]
                    return []

                def arrow_body(self, items):
                    """arrow_body: assignment_expr | block_stmt"""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Expr):
                            return [item]
                        elif isinstance(item, list):
                            return item
                    return []

                # Template Literals
                def template_literal(self, items):
                    """template_literal: TEMPLATE_STRING | template_head template_spans"""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Token):
                            return uni.String(value=item.value, kid=[item])
                        return item if isinstance(item, uni.Expr) else None
                    return None

                # Types
                def type(self, items):
                    """type: union_type | intersection_type | primary_type | ..."""
                    return (
                        items[0] if items and isinstance(items[0], uni.Expr) else None
                    )

                def union_type(self, items):
                    """union_type: type (BW_OR type)+"""
                    types = [i for i in items if isinstance(i, uni.Expr)]
                    if len(types) == 1:
                        return types[0]
                    # Return as tuple for union representation
                    return uni.TupleVal(values=types, kid=list(items))

                def intersection_type(self, items):
                    """intersection_type: type (BW_AND type)+"""
                    types = [i for i in items if isinstance(i, uni.Expr)]
                    if len(types) == 1:
                        return types[0]
                    return uni.TupleVal(values=types, kid=list(items))

                def primary_type(self, items):
                    """primary_type: predefined_type | type_reference | ..."""
                    return (
                        items[0] if items and isinstance(items[0], uni.Expr) else None
                    )

                def predefined_type(self, items):
                    """predefined_type: KW_ANY | KW_STRING_TYPE | ..."""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Token):
                            return uni.BuiltinType(
                                typ=item,
                                kid=[item],
                            )
                    return None

                def type_reference(self, items):
                    """type_reference: type_name type_args?"""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Name):
                            return item
                        elif isinstance(item, uni.Token) and item.name in (
                            Tok.NAME.name,
                            "NAME",
                        ):
                            return self._make_name(item)
                    return None

                def type_name(self, items):
                    """type_name: NAME (DOT NAME)*"""
                    names = []
                    for item in items:
                        if isinstance(item, uni.Token) and item.name in (
                            Tok.NAME.name,
                            "NAME",
                        ):
                            names.append(self._make_name(item))
                        elif isinstance(item, uni.Name):
                            names.append(item)
                    return names[0] if names else None

                # JSX
                def jsx_element(self, items):
                    """jsx_element: jsx_self_closing | jsx_opening_closing | jsx_fragment"""
                    return items[0] if items else None

                def jsx_self_closing(self, items):
                    """jsx_self_closing: JSX_OPEN_START jsx_element_name jsx_attributes? JSX_SELF_CLOSE"""
                    return uni.JsxElement(
                        tag=items[1] if len(items) > 1 else None,
                        props=[],
                        children=[],
                        kid=list(items),
                    )

                def jsx_opening_closing(self, items):
                    """jsx_opening_closing: jsx_opening_element jsx_children? jsx_closing_element"""
                    return uni.JsxElement(
                        tag=items[0] if items else None,
                        props=[],
                        children=[],
                        kid=list(items),
                    )

                def jsx_fragment(self, items):
                    """jsx_fragment: JSX_FRAG_OPEN jsx_children? JSX_FRAG_CLOSE"""
                    return uni.JsxElement(
                        tag=None,
                        props=[],
                        children=[],
                        kid=list(items),
                    )

                def jsx_element_name(self, items):
                    """jsx_element_name: NAME (DOT NAME)*"""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Token):
                            return self._make_name(item)
                        return item
                    return None

                # Decorators
                def decorators(self, items):
                    """decorators: decorator+"""
                    return [i for i in items if isinstance(i, uni.Expr)]

                def decorator(self, items):
                    """decorator: AT decorator_expr"""
                    for item in items:
                        if isinstance(item, uni.Expr):
                            return item
                    return None

                def decorator_expr(self, items):
                    """decorator_expr: NAME | decorator_expr DOT NAME | decorator_expr arguments"""
                    if items:
                        item = items[0]
                        if isinstance(item, uni.Token) and item.name in (
                            Tok.NAME.name,
                            "NAME",
                        ):
                            return self._make_name(item)
                        return item if isinstance(item, uni.Expr) else None
                    return None

                # Default handler
                def __default__(self, data, children, meta):
                    """Default handler for unhandled rules."""
                    if children:
                        # Return first non-None child
                        for child in children:
                            if child is not None:
                                return child
                    return None

            transformer = _InnerTransformer()
            return transformer.transform(tree)

        def _make_module(self, body: list[uni.UniNode]) -> uni.Module:
            """Create a Module node."""
            stmts = [
                s for s in body if isinstance(s, (uni.ElementStmt, uni.CodeBlockStmt))
            ]
            mod_name = os.path.basename(self.parse_ref.mod_path)
            if mod_name.endswith(".ts"):
                mod_name = mod_name[:-3]
            elif mod_name.endswith(".tsx"):
                mod_name = mod_name[:-4]
            elif mod_name.endswith(".js"):
                mod_name = mod_name[:-3]
            elif mod_name.endswith(".jsx"):
                mod_name = mod_name[:-4]

            # Cast stmts since TypeScript parser produces CodeBlockStmt which
            # is not in the standard Module.body type but works at runtime
            return uni.Module(
                name=mod_name,
                source=self.parse_ref.ir_in,
                doc=None,
                body=cast(list[uni.ElementStmt | uni.String | uni.EmptyToken], stmts),
                terminals=self.terminals,
                kid=body or [uni.EmptyToken(self.parse_ref.ir_in)],
            )

        def ice(self) -> Exception:
            """Raise internal compiler error."""
            self.parse_ref.log_error("Internal Compiler Error, Invalid Parse Tree!")
            return RuntimeError(
                f"{self.parse_ref.__class__.__name__} - Internal Compiler Error!"
            )



from typing import Final

from . import affinity as affinity
from ._coverage import *
from ._geometry import *
from .constructive import *
from .coordinates import *
from .creation import *
from .errors import setup_signal_checks as setup_signal_checks
from .geometry import (
    GeometryCollection as GeometryCollection,
    LinearRing as LinearRing,
    LineString as LineString,
    MultiLineString as MultiLineString,
    MultiPoint as MultiPoint,
    MultiPolygon as MultiPolygon,
    Point as Point,
    Polygon as Polygon,
)
from .io import *
from .lib import (
    Geometry as Geometry,
    GEOSException as GEOSException,
    geos_capi_version as geos_capi_version,
    geos_capi_version_string as geos_capi_version_string,
    geos_version as geos_version,
    geos_version_string as geos_version_string,
)
from .linear import *
from .measurement import *
from .predicates import *
from .set_operations import *
from .strtree import *

__version__: Final[str]



import sys
import types
import unittest
from _typeshed import ExcInfo
from collections.abc import Callable
from typing import Any, Final, NamedTuple, type_check_only
from typing_extensions import Self, TypeAlias

__all__ = [
    "register_optionflag",
    "DONT_ACCEPT_TRUE_FOR_1",
    "DONT_ACCEPT_BLANKLINE",
    "NORMALIZE_WHITESPACE",
    "ELLIPSIS",
    "SKIP",
    "IGNORE_EXCEPTION_DETAIL",
    "COMPARISON_FLAGS",
    "REPORT_UDIFF",
    "REPORT_CDIFF",
    "REPORT_NDIFF",
    "REPORT_ONLY_FIRST_FAILURE",
    "REPORTING_FLAGS",
    "FAIL_FAST",
    "Example",
    "DocTest",
    "DocTestParser",
    "DocTestFinder",
    "DocTestRunner",
    "OutputChecker",
    "DocTestFailure",
    "UnexpectedException",
    "DebugRunner",
    "testmod",
    "testfile",
    "run_docstring_examples",
    "DocTestSuite",
    "DocFileSuite",
    "set_unittest_reportflags",
    "script_from_examples",
    "testsource",
    "debug_src",
    "debug",
]

if sys.version_info >= (3, 13):
    @type_check_only
    class _TestResultsBase(NamedTuple):
        failed: int
        attempted: int

    class TestResults(_TestResultsBase):
        def __new__(cls, failed: int, attempted: int, *, skipped: int = 0) -> Self: ...
        skipped: int

else:
    class TestResults(NamedTuple):
        failed: int
        attempted: int

OPTIONFLAGS_BY_NAME: Final[dict[str, int]]

def register_optionflag(name: str) -> int: ...

DONT_ACCEPT_TRUE_FOR_1: Final = 1
DONT_ACCEPT_BLANKLINE: Final = 2
NORMALIZE_WHITESPACE: Final = 4
ELLIPSIS: Final = 8
SKIP: Final = 16
IGNORE_EXCEPTION_DETAIL: Final = 32

COMPARISON_FLAGS: Final = 63

REPORT_UDIFF: Final = 64
REPORT_CDIFF: Final = 128
REPORT_NDIFF: Final = 256
REPORT_ONLY_FIRST_FAILURE: Final = 512
FAIL_FAST: Final = 1024

REPORTING_FLAGS: Final = 1984

BLANKLINE_MARKER: Final = "<BLANKLINE>"
ELLIPSIS_MARKER: Final = "..."

class Example:
    source: str
    want: str
    exc_msg: str | None
    lineno: int
    indent: int
    options: dict[int, bool]
    def __init__(
        self,
        source: str,
        want: str,
        exc_msg: str | None = None,
        lineno: int = 0,
        indent: int = 0,
        options: dict[int, bool] | None = None,
    ) -> None: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...

class DocTest:
    examples: list[Example]
    globs: dict[str, Any]
    name: str
    filename: str | None
    lineno: int | None
    docstring: str | None
    def __init__(
        self,
        examples: list[Example],
        globs: dict[str, Any],
        name: str,
        filename: str | None,
        lineno: int | None,
        docstring: str | None,
    ) -> None: ...
    def __hash__(self) -> int: ...
    def __lt__(self, other: DocTest) -> bool: ...
    def __eq__(self, other: object) -> bool: ...

class DocTestParser:
    def parse(self, string: str, name: str = "<string>") -> list[str | Example]: ...
    def get_doctest(self, string: str, globs: dict[str, Any], name: str, filename: str | None, lineno: int | None) -> DocTest: ...
    def get_examples(self, string: str, name: str = "<string>") -> list[Example]: ...

class DocTestFinder:
    def __init__(
        self, verbose: bool = False, parser: DocTestParser = ..., recurse: bool = True, exclude_empty: bool = True
    ) -> None: ...
    def find(
        self,
        obj: object,
        name: str | None = None,
        module: None | bool | types.ModuleType = None,
        globs: dict[str, Any] | None = None,
        extraglobs: dict[str, Any] | None = None,
    ) -> list[DocTest]: ...

_Out: TypeAlias = Callable[[str], object]

class DocTestRunner:
    DIVIDER: str
    optionflags: int
    original_optionflags: int
    tries: int
    failures: int
    if sys.version_info >= (3, 13):
        skips: int
    test: DocTest
    def __init__(self, checker: OutputChecker | None = None, verbose: bool | None = None, optionflags: int = 0) -> None: ...
    def report_start(self, out: _Out, test: DocTest, example: Example) -> None: ...
    def report_success(self, out: _Out, test: DocTest, example: Example, got: str) -> None: ...
    def report_failure(self, out: _Out, test: DocTest, example: Example, got: str) -> None: ...
    def report_unexpected_exception(self, out: _Out, test: DocTest, example: Example, exc_info: ExcInfo) -> None: ...
    def run(
        self, test: DocTest, compileflags: int | None = None, out: _Out | None = None, clear_globs: bool = True
    ) -> TestResults: ...
    def summarize(self, verbose: bool | None = None) -> TestResults: ...
    def merge(self, other: DocTestRunner) -> None: ...

class OutputChecker:
    def check_output(self, want: str, got: str, optionflags: int) -> bool: ...
    def output_difference(self, example: Example, got: str, optionflags: int) -> str: ...

class DocTestFailure(Exception):
    test: DocTest
    example: Example
    got: str
    def __init__(self, test: DocTest, example: Example, got: str) -> None: ...

class UnexpectedException(Exception):
    test: DocTest
    example: Example
    exc_info: ExcInfo
    def __init__(self, test: DocTest, example: Example, exc_info: ExcInfo) -> None: ...

class DebugRunner(DocTestRunner): ...

master: DocTestRunner | None

def testmod(
    m: types.ModuleType | None = None,
    name: str | None = None,
    globs: dict[str, Any] | None = None,
    verbose: bool | None = None,
    report: bool = True,
    optionflags: int = 0,
    extraglobs: dict[str, Any] | None = None,
    raise_on_error: bool = False,
    exclude_empty: bool = False,
) -> TestResults: ...
def testfile(
    filename: str,
    module_relative: bool = True,
    name: str | None = None,
    package: None | str | types.ModuleType = None,
    globs: dict[str, Any] | None = None,
    verbose: bool | None = None,
    report: bool = True,
    optionflags: int = 0,
    extraglobs: dict[str, Any] | None = None,
    raise_on_error: bool = False,
    parser: DocTestParser = ...,
    encoding: str | None = None,
) -> TestResults: ...
def run_docstring_examples(
    f: object,
    globs: dict[str, Any],
    verbose: bool = False,
    name: str = "NoName",
    compileflags: int | None = None,
    optionflags: int = 0,
) -> None: ...
def set_unittest_reportflags(flags: int) -> int: ...

class DocTestCase(unittest.TestCase):
    def __init__(
        self,
        test: DocTest,
        optionflags: int = 0,
        setUp: Callable[[DocTest], object] | None = None,
        tearDown: Callable[[DocTest], object] | None = None,
        checker: OutputChecker | None = None,
    ) -> None: ...
    def runTest(self) -> None: ...
    def format_failure(self, err: str) -> str: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...

class SkipDocTestCase(DocTestCase):
    def __init__(self, module: types.ModuleType) -> None: ...
    def test_skip(self) -> None: ...

class _DocTestSuite(unittest.TestSuite): ...

def DocTestSuite(
    module: None | str | types.ModuleType = None,
    globs: dict[str, Any] | None = None,
    extraglobs: dict[str, Any] | None = None,
    test_finder: DocTestFinder | None = None,
    **options: Any,
) -> _DocTestSuite: ...

class DocFileCase(DocTestCase): ...

def DocFileTest(
    path: str,
    module_relative: bool = True,
    package: None | str | types.ModuleType = None,
    globs: dict[str, Any] | None = None,
    parser: DocTestParser = ...,
    encoding: str | None = None,
    **options: Any,
) -> DocFileCase: ...
def DocFileSuite(*paths: str, **kw: Any) -> _DocTestSuite: ...
def script_from_examples(s: str) -> str: ...
def testsource(module: None | str | types.ModuleType, name: str) -> str: ...
def debug_src(src: str, pm: bool = False, globs: dict[str, Any] | None = None) -> None: ...
def debug_script(src: str, pm: bool = False, globs: dict[str, Any] | None = None) -> None: ...
def debug(module: None | str | types.ModuleType, name: str, pm: bool = False) -> None: ...



from hvac.api.vault_api_base import VaultApiBase

DEFAULT_MOUNT_POINT: str

class ActiveDirectory(VaultApiBase):
    def configure(
        self,
        binddn=None,
        bindpass=None,
        url=None,
        userdn=None,
        upndomain=None,
        ttl=None,
        max_ttl=None,
        mount_point="ad",
        *args,
        **kwargs,
    ): ...
    def read_config(self, mount_point="ad"): ...
    def create_or_update_role(self, name, service_account_name=None, ttl=None, mount_point="ad"): ...
    def read_role(self, name, mount_point="ad"): ...
    def list_roles(self, mount_point="ad"): ...
    def delete_role(self, name, mount_point="ad"): ...
    def generate_credentials(self, name, mount_point="ad"): ...



from collections.abc import Callable, Mapping
from hashlib import _Hash
from typing import Final

from .backends.base import Key

class Algorithms:
    NONE: str
    HS256: str
    HS384: str
    HS512: str
    RS256: str
    RS384: str
    RS512: str
    ES256: str
    ES384: str
    ES512: str
    A128CBC_HS256: str
    A192CBC_HS384: str
    A256CBC_HS512: str
    A128GCM: str
    A192GCM: str
    A256GCM: str
    A128CBC: str
    A192CBC: str
    A256CBC: str
    DIR: str
    RSA1_5: str
    RSA_OAEP: str
    RSA_OAEP_256: str
    A128KW: str
    A192KW: str
    A256KW: str
    ECDH_ES: str
    ECDH_ES_A128KW: str
    ECDH_ES_A192KW: str
    ECDH_ES_A256KW: str
    A128GCMKW: str
    A192GCMKW: str
    A256GCMKW: str
    PBES2_HS256_A128KW: str
    PBES2_HS384_A192KW: str
    PBES2_HS512_A256KW: str
    DEF: str
    HMAC: set[str]
    RSA_DS: set[str]
    RSA_KW: set[str]
    RSA: set[str]
    EC_DS: set[str]
    EC_KW: set[str]
    EC: set[str]
    AES_PSEUDO: set[str]
    AES_JWE_ENC: set[str]
    AES_ENC: set[str]
    AES_KW: set[str]
    AEC_GCM_KW: set[str]
    AES: set[str]
    PBES2_KW: set[str]
    HMAC_AUTH_TAG: set[str]
    GCM: set[str]
    SUPPORTED: set[str]
    ALL: set[str]
    HASHES: Mapping[str, Callable[[bytes], _Hash]]
    KEYS: Mapping[str, type[Key]]

ALGORITHMS: Algorithms

class Zips:
    DEF: str
    NONE: None
    SUPPORTED: set[str | None]

ZIPS: Zips

JWE_SIZE_LIMIT: Final[int]



import datetime
from _typeshed import Incomplete, StrPath
from collections.abc import Generator, Iterable
from typing import ClassVar, Final, Literal, TypeVar, overload

_D = TypeVar("_D", datetime.date, datetime.datetime)
_DT = TypeVar("_DT", datetime.date, datetime.datetime, datetime.timedelta)

MON: Final = 0
TUE: Final = 1
WED: Final = 2
THU: Final = 3
FRI: Final = 4
SAT: Final = 5
SUN: Final = 6
ISO_MON: Final = 1
ISO_TUE: Final = 2
ISO_WED: Final = 3
ISO_THU: Final = 4
ISO_FRI: Final = 5
ISO_SAT: Final = 6
ISO_SUN: Final = 7

@overload
def cleaned_date(day: _D, keep_datetime: Literal[True]) -> _D: ...
@overload
def cleaned_date(day: datetime.date | datetime.datetime, keep_datetime: Literal[False] | None = False) -> datetime.date: ...
def daterange(start: _DT, end: _DT) -> Generator[_DT]: ...

class ChristianMixin:
    EASTER_METHOD: ClassVar[Literal[1, 2, 3] | None]
    include_epiphany: ClassVar[bool]
    include_clean_monday: ClassVar[bool]
    include_annunciation: ClassVar[bool]
    include_fat_tuesday: ClassVar[bool]
    fat_tuesday_label: ClassVar[str | None]
    include_ash_wednesday: ClassVar[bool]
    ash_wednesday_label: ClassVar[str]
    include_palm_sunday: ClassVar[bool]
    include_holy_thursday: ClassVar[bool]
    holy_thursday_label: ClassVar[str]
    include_good_friday: ClassVar[bool]
    good_friday_label: ClassVar[str]
    include_easter_monday: ClassVar[bool]
    include_easter_saturday: ClassVar[bool]
    easter_saturday_label: ClassVar[str]
    include_easter_sunday: ClassVar[bool]
    include_all_saints: ClassVar[bool]
    include_immaculate_conception: ClassVar[bool]
    immaculate_conception_label: ClassVar[str]
    include_christmas: ClassVar[bool]
    christmas_day_label: ClassVar[str]
    include_christmas_eve: ClassVar[bool]
    include_ascension: ClassVar[bool]
    include_assumption: ClassVar[bool]
    include_whit_sunday: ClassVar[bool]
    whit_sunday_label: ClassVar[str]
    include_whit_monday: ClassVar[bool]
    whit_monday_label: ClassVar[str]
    include_corpus_christi: ClassVar[bool]
    include_boxing_day: ClassVar[bool]
    boxing_day_label: ClassVar[str]
    include_all_souls: ClassVar[bool]
    def get_fat_tuesday(self, year: int) -> datetime.date: ...
    def get_ash_wednesday(self, year: int) -> datetime.date: ...
    def get_palm_sunday(self, year: int) -> datetime.date: ...
    def get_holy_thursday(self, year: int) -> datetime.date: ...
    def get_good_friday(self, year: int) -> datetime.date: ...
    def get_clean_monday(self, year: int) -> datetime.date: ...
    def get_easter_saturday(self, year: int) -> datetime.date: ...
    def get_easter_sunday(self, year: int) -> datetime.date: ...
    def get_easter_monday(self, year: int) -> datetime.date: ...
    def get_ascension_thursday(self, year: int) -> datetime.date: ...
    def get_whit_monday(self, year: int) -> datetime.date: ...
    def get_whit_sunday(self, year: int) -> datetime.date: ...
    def get_corpus_christi(self, year: int) -> datetime.date: ...
    def shift_christmas_boxing_days(self, year: int) -> list[tuple[datetime.date, str]]: ...
    def get_variable_days(self, year: int) -> list[tuple[datetime.date, str]]: ...

class WesternMixin(ChristianMixin):
    EASTER_METHOD: ClassVar[Literal[1, 2, 3]] = 3
    WEEKEND_DAYS: ClassVar[tuple[int, ...]]

class OrthodoxMixin(ChristianMixin):
    EASTER_METHOD: ClassVar[Literal[1, 2, 3]] = 2
    WEEKEND_DAYS: ClassVar[tuple[int, ...]]
    include_orthodox_christmas: ClassVar[bool]
    orthodox_christmas_day_label: ClassVar[str]
    def get_fixed_holidays(self, year: int) -> list[tuple[datetime.date, str]]: ...

class LunarMixin:
    @staticmethod
    def lunar(year: int, month: int, day: int) -> datetime.date: ...

class ChineseNewYearMixin(LunarMixin):
    include_chinese_new_year_eve: ClassVar[bool]
    chinese_new_year_eve_label: ClassVar[str]
    include_chinese_new_year: ClassVar[bool]
    chinese_new_year_label: ClassVar[str]
    include_chinese_second_day: ClassVar[bool]
    chinese_second_day_label: ClassVar[str]
    include_chinese_third_day: ClassVar[bool]
    chinese_third_day_label: ClassVar[str]
    shift_sunday_holidays: ClassVar[bool]
    shift_start_cny_sunday: ClassVar[bool]
    def get_chinese_new_year(self, year: int) -> list[tuple[datetime.date, str]]: ...
    def get_variable_days(self, year: int) -> list[tuple[datetime.date, str]]: ...
    def get_shifted_holidays(self, dates: Iterable[tuple[_D, str]]) -> Generator[tuple[_D, str]]: ...
    def get_calendar_holidays(self, year: int) -> list[tuple[datetime.date, str]]: ...

class CalverterMixin:
    conversion_method: Incomplete
    ISLAMIC_HOLIDAYS: ClassVar[tuple[tuple[int, int, str], ...]]
    def __init__(self, *args, **kwargs) -> None: ...
    def converted(self, year: int) -> list[tuple[int, int, int]]: ...
    def calverted_years(self, year: int) -> list[int]: ...
    def get_islamic_holidays(self) -> tuple[tuple[int, int, str], ...]: ...
    def get_delta_islamic_holidays(self, year: int) -> datetime.timedelta | None: ...
    def get_variable_days(self, year: int) -> list[tuple[datetime.date, str]]: ...

class IslamicMixin(CalverterMixin):
    WEEKEND_DAYS: ClassVar[tuple[int, ...]]
    conversion_method: Incomplete
    include_prophet_birthday: ClassVar[bool]
    include_day_after_prophet_birthday: ClassVar[bool]
    include_start_ramadan: ClassVar[bool]
    include_eid_al_fitr: ClassVar[bool]
    length_eid_al_fitr: int
    eid_al_fitr_label: ClassVar[str]
    include_eid_al_adha: ClassVar[bool]
    eid_al_adha_label: ClassVar[str]
    length_eid_al_adha: int
    include_day_of_sacrifice: ClassVar[bool]
    day_of_sacrifice_label: ClassVar[str]
    include_islamic_new_year: ClassVar[bool]
    include_laylat_al_qadr: ClassVar[bool]
    include_nuzul_al_quran: ClassVar[bool]
    def get_islamic_holidays(self) -> tuple[tuple[int, int, str]]: ...

class CoreCalendar:
    FIXED_HOLIDAYS: ClassVar[tuple[tuple[int, int, str], ...]]
    WEEKEND_DAYS: ClassVar[tuple[int, ...]]
    def __init__(self) -> None: ...
    def name(cls) -> str: ...
    def get_fixed_holidays(self, year: int) -> list[tuple[datetime.date, str]]: ...
    def get_variable_days(self, year: int) -> list[tuple[datetime.date, str]]: ...
    def get_calendar_holidays(self, year: int) -> list[tuple[datetime.date, str]]: ...
    def holidays(self, year: int | None = None) -> list[tuple[datetime.date, str]]: ...
    def get_holiday_label(self, day: datetime.date | datetime.datetime) -> str | None: ...
    def holidays_set(self, year: int | None = None) -> set[datetime.date]: ...
    def get_weekend_days(self) -> tuple[int, ...]: ...
    def is_working_day(
        self,
        day: datetime.date | datetime.datetime,
        extra_working_days: Iterable[datetime.date | datetime.datetime] | None = None,
        extra_holidays: Iterable[datetime.date | datetime.datetime] | None = None,
    ) -> bool: ...
    def is_holiday(
        self, day: datetime.date | datetime.datetime, extra_holidays: Iterable[datetime.date | datetime.datetime] | None = None
    ) -> bool: ...
    def add_working_days(
        self,
        day: datetime.date | datetime.datetime,
        delta: int,
        extra_working_days: Iterable[datetime.date | datetime.datetime] | None = None,
        extra_holidays: Iterable[datetime.date | datetime.datetime] | None = None,
        keep_datetime: bool = False,
    ) -> datetime.date: ...
    def sub_working_days(
        self,
        day: datetime.date | datetime.datetime,
        delta: int,
        extra_working_days: Iterable[datetime.date | datetime.datetime] | None = None,
        extra_holidays: Iterable[datetime.date | datetime.datetime] | None = None,
        keep_datetime: bool = False,
    ) -> datetime.date: ...
    def find_following_working_day(self, day: datetime.date) -> datetime.date: ...
    @staticmethod
    def get_nth_weekday_in_month(
        year: int, month: int, weekday: int, n: int = 1, start: datetime.date | Literal[False] | None = None
    ) -> datetime.date | None: ...
    @staticmethod
    def get_last_weekday_in_month(year: int, month: int, weekday: int) -> datetime.date: ...
    @staticmethod
    def get_iso_week_date(year: int, week_nb: int, weekday: int = 1) -> datetime.date: ...
    @staticmethod
    def get_first_weekday_after(day: _D, weekday: int) -> _D: ...
    def get_working_days_delta(
        self,
        start: datetime.date | datetime.datetime,
        end: datetime.date | datetime.datetime,
        include_start: bool = False,
        extra_working_days: Iterable[datetime.date | datetime.datetime] | None = None,
        extra_holidays: Iterable[datetime.date | datetime.datetime] | None = None,
    ) -> int: ...
    def export_to_ical(
        self, period: tuple[int, int] | list[int] = [2000, 2030], target_path: StrPath | None = None
    ) -> str | None: ...

class Calendar(CoreCalendar):
    include_new_years_day: ClassVar[bool]
    include_new_years_eve: ClassVar[bool]
    shift_new_years_day: ClassVar[bool]
    include_labour_day: ClassVar[bool]
    labour_day_label: ClassVar[str]
    def __init__(self, **kwargs) -> None: ...
    def get_fixed_holidays(self, year: int) -> list[tuple[datetime.date, str]]: ...
    def get_variable_days(self, year: int) -> list[tuple[datetime.date, str]]: ...

class WesternCalendar(WesternMixin, Calendar): ...
class OrthodoxCalendar(OrthodoxMixin, Calendar): ...

class ChineseNewYearCalendar(ChineseNewYearMixin, Calendar):
    WEEKEND_DAYS: ClassVar[tuple[int, ...]]

class IslamicCalendar(IslamicMixin, Calendar): ...

class IslamoWesternCalendar(IslamicMixin, WesternMixin, Calendar):
    FIXED_HOLIDAYS: ClassVar[tuple[tuple[int, int, str], ...]]



from _typeshed import Incomplete
from collections.abc import Callable

def decorate_all_functions(function_decorator: Callable[..., Incomplete]) -> Callable[..., Incomplete]: ...
def xray_on_call(cls, func: Callable[..., Incomplete]) -> Callable[..., Incomplete]: ...
def parse_bind(bind) -> dict[str, str]: ...



from .__meta__ import __version__ as __version__, __version_info__ as __version_info__
from .core import Markdown as Markdown, markdown as markdown, markdownFromFile as markdownFromFile
from .extensions import Extension as Extension

__all__ = ["Markdown", "markdown", "markdownFromFile"]



from _typeshed import SupportsRead, SupportsWrite
from collections.abc import Callable
from typing import Any, Final

__version__: Final[str]

def encode(
    obj: Any,
    ensure_ascii: bool = ...,
    double_precision: int = ...,
    encode_html_chars: bool = ...,
    escape_forward_slashes: bool = ...,
    sort_keys: bool = ...,
    indent: int = ...,
    allow_nan: bool = ...,
    reject_bytes: bool = ...,
    default: Callable[[Any], Any] | None = None,  # Specify how to serialize arbitrary types
    separators: tuple[str, str] | None = None,
) -> str: ...
def dumps(
    obj: Any,
    ensure_ascii: bool = ...,
    double_precision: int = ...,
    encode_html_chars: bool = ...,
    escape_forward_slashes: bool = ...,
    sort_keys: bool = ...,
    indent: int = ...,
    allow_nan: bool = ...,
    reject_bytes: bool = ...,
    default: Callable[[Any], Any] | None = None,  # Specify how to serialize arbitrary types
    separators: tuple[str, str] | None = None,
) -> str: ...
def dump(
    obj: Any,
    fp: SupportsWrite[str],
    *,
    ensure_ascii: bool = ...,
    double_precision: int = ...,
    encode_html_chars: bool = ...,
    escape_forward_slashes: bool = ...,
    sort_keys: bool = ...,
    indent: int = ...,
    allow_nan: bool = ...,
    reject_bytes: bool = ...,
    default: Callable[[Any], Any] | None = None,  # Specify how to serialize arbitrary types
    separators: tuple[str, str] | None = None,
) -> None: ...
def decode(s: str | bytes | bytearray, precise_float: bool = ...) -> Any: ...
def loads(s: str | bytes | bytearray, precise_float: bool = ...) -> Any: ...
def load(fp: SupportsRead[str | bytes | bytearray], precise_float: bool = ...) -> Any: ...

class JSONDecodeError(ValueError): ...



from collections.abc import Iterator
from typing import Any

from matplotlib.axes import Axes
from matplotlib.figure import Figure, SubFigure
from seaborn._core.plot import FacetSpec, PairSpec

class Subplots:
    subplot_spec: dict[str, Any]
    def __init__(self, subplot_spec: dict[str, Any], facet_spec: FacetSpec, pair_spec: PairSpec) -> None: ...
    def init_figure(
        self,
        pair_spec: PairSpec,
        pyplot: bool = False,
        figure_kws: dict[str, Any] | None = None,
        target: Axes | Figure | SubFigure | None = None,
    ) -> Figure: ...
    def __iter__(self) -> Iterator[dict[str, Any]]: ...
    def __len__(self) -> int: ...



from __future__ import annotations

import logging
import logging.handlers
import multiprocessing
import queue
from typing import Any

# This pattern comes from the logging docs, and should therefore pass a type checker
# See https://docs.python.org/3/library/logging.html#logrecord-objects

old_factory = logging.getLogRecordFactory()


def record_factory(*args: Any, **kwargs: Any) -> logging.LogRecord:
    record = old_factory(*args, **kwargs)
    record.custom_attribute = 0xDECAFBAD
    return record


logging.setLogRecordFactory(record_factory)

# The logging docs say that QueueHandler and QueueListener can take "any queue-like object"
# We test that here (regression test for #10168)
logging.handlers.QueueHandler(queue.Queue())
logging.handlers.QueueHandler(queue.SimpleQueue())
logging.handlers.QueueHandler(multiprocessing.Queue())
logging.handlers.QueueListener(queue.Queue())
logging.handlers.QueueListener(queue.SimpleQueue())
logging.handlers.QueueListener(multiprocessing.Queue())



from ..core import WesternCalendar

class Spain(WesternCalendar): ...
class Andalusia(Spain): ...
class Aragon(Spain): ...
class CastileAndLeon(Spain): ...
class CastillaLaMancha(Spain): ...
class CanaryIslands(Spain): ...
class Catalonia(Spain): ...
class Extremadura(Spain): ...
class Galicia(Spain): ...
class BalearicIslands(Spain): ...
class LaRioja(Spain): ...
class CommunityofMadrid(Spain): ...
class Murcia(Spain): ...
class Navarre(Spain): ...
class Asturias(Spain): ...
class BasqueCountry(Spain): ...
class Cantabria(Spain): ...
class ValencianCommunity(Spain): ...



from hvac.api.system_backend.audit import Audit as Audit
from hvac.api.system_backend.auth import Auth as Auth
from hvac.api.system_backend.capabilities import Capabilities as Capabilities
from hvac.api.system_backend.health import Health as Health
from hvac.api.system_backend.init import Init as Init
from hvac.api.system_backend.key import Key as Key
from hvac.api.system_backend.leader import Leader as Leader
from hvac.api.system_backend.lease import Lease as Lease
from hvac.api.system_backend.mount import Mount as Mount
from hvac.api.system_backend.namespace import Namespace as Namespace
from hvac.api.system_backend.policies import Policies as Policies
from hvac.api.system_backend.policy import Policy as Policy
from hvac.api.system_backend.quota import Quota as Quota
from hvac.api.system_backend.raft import Raft as Raft
from hvac.api.system_backend.seal import Seal as Seal
from hvac.api.system_backend.system_backend_mixin import SystemBackendMixin as SystemBackendMixin
from hvac.api.system_backend.wrapping import Wrapping as Wrapping
from hvac.api.vault_api_base import VaultApiBase
from hvac.api.vault_api_category import VaultApiCategory

__all__ = (
    "Audit",
    "Auth",
    "Capabilities",
    "Health",
    "Init",
    "Key",
    "Leader",
    "Lease",
    "Mount",
    "Namespace",
    "Policies",
    "Policy",
    "Quota",
    "Raft",
    "Seal",
    "SystemBackend",
    "SystemBackendMixin",
    "Wrapping",
)

class SystemBackend(
    VaultApiCategory,
    Audit,
    Auth,
    Capabilities,
    Health,
    Init,
    Key,
    Leader,
    Lease,
    Mount,
    Namespace,
    Policies,
    Policy,
    Quota,
    Raft,
    Seal,
    Wrapping,
):
    implemented_classes: list[type[VaultApiBase]]
    unimplemented_classes: list[str]



from braintree.ach_mandate import AchMandate as AchMandate
from braintree.add_on import AddOn as AddOn
from braintree.add_on_gateway import AddOnGateway as AddOnGateway
from braintree.address import Address as Address
from braintree.address_gateway import AddressGateway as AddressGateway
from braintree.amex_express_checkout_card import AmexExpressCheckoutCard as AmexExpressCheckoutCard
from braintree.android_pay_card import AndroidPayCard as AndroidPayCard
from braintree.apple_pay_card import ApplePayCard as ApplePayCard
from braintree.apple_pay_gateway import ApplePayGateway as ApplePayGateway
from braintree.bank_account_instant_verification_gateway import (
    BankAccountInstantVerificationGateway as BankAccountInstantVerificationGateway,
)
from braintree.bank_account_instant_verification_jwt import BankAccountInstantVerificationJwt as BankAccountInstantVerificationJwt
from braintree.bank_account_instant_verification_jwt_request import (
    BankAccountInstantVerificationJwtRequest as BankAccountInstantVerificationJwtRequest,
)
from braintree.blik_alias import BlikAlias as BlikAlias
from braintree.braintree_gateway import BraintreeGateway as BraintreeGateway
from braintree.client_token import ClientToken as ClientToken
from braintree.configuration import Configuration as Configuration
from braintree.connected_merchant_paypal_status_changed import (
    ConnectedMerchantPayPalStatusChanged as ConnectedMerchantPayPalStatusChanged,
)
from braintree.connected_merchant_status_transitioned import (
    ConnectedMerchantStatusTransitioned as ConnectedMerchantStatusTransitioned,
)
from braintree.credentials_parser import CredentialsParser as CredentialsParser
from braintree.credit_card import CreditCard as CreditCard
from braintree.credit_card_gateway import CreditCardGateway as CreditCardGateway
from braintree.credit_card_verification import CreditCardVerification as CreditCardVerification
from braintree.credit_card_verification_search import CreditCardVerificationSearch as CreditCardVerificationSearch
from braintree.customer import Customer as Customer
from braintree.customer_gateway import CustomerGateway as CustomerGateway
from braintree.customer_search import CustomerSearch as CustomerSearch
from braintree.descriptor import Descriptor as Descriptor
from braintree.disbursement import Disbursement as Disbursement
from braintree.discount import Discount as Discount
from braintree.discount_gateway import DiscountGateway as DiscountGateway
from braintree.dispute import Dispute as Dispute
from braintree.dispute_search import DisputeSearch as DisputeSearch
from braintree.document_upload import DocumentUpload as DocumentUpload
from braintree.document_upload_gateway import DocumentUploadGateway as DocumentUploadGateway
from braintree.enriched_customer_data import EnrichedCustomerData as EnrichedCustomerData
from braintree.environment import Environment as Environment
from braintree.error_codes import ErrorCodes as ErrorCodes
from braintree.error_result import ErrorResult as ErrorResult
from braintree.errors import Errors as Errors
from braintree.europe_bank_account import EuropeBankAccount as EuropeBankAccount
from braintree.graphql import *
from braintree.liability_shift import LiabilityShift as LiabilityShift
from braintree.local_payment_completed import LocalPaymentCompleted as LocalPaymentCompleted
from braintree.local_payment_reversed import LocalPaymentReversed as LocalPaymentReversed
from braintree.merchant import Merchant as Merchant
from braintree.merchant_account import MerchantAccount as MerchantAccount
from braintree.merchant_account_gateway import MerchantAccountGateway as MerchantAccountGateway
from braintree.monetary_amount import MonetaryAmount as MonetaryAmount
from braintree.oauth_access_revocation import OAuthAccessRevocation as OAuthAccessRevocation
from braintree.partner_merchant import PartnerMerchant as PartnerMerchant
from braintree.payment_facilitator import PaymentFacilitator as PaymentFacilitator
from braintree.payment_instrument_type import PaymentInstrumentType as PaymentInstrumentType
from braintree.payment_method import PaymentMethod as PaymentMethod
from braintree.payment_method_customer_data_updated_metadata import (
    PaymentMethodCustomerDataUpdatedMetadata as PaymentMethodCustomerDataUpdatedMetadata,
)
from braintree.payment_method_nonce import PaymentMethodNonce as PaymentMethodNonce
from braintree.payment_method_parser import parse_payment_method as parse_payment_method
from braintree.paypal_account import PayPalAccount as PayPalAccount
from braintree.paypal_payment_resource import PayPalPaymentResource as PayPalPaymentResource
from braintree.plan import Plan as Plan
from braintree.plan_gateway import PlanGateway as PlanGateway
from braintree.processor_response_types import ProcessorResponseTypes as ProcessorResponseTypes
from braintree.receiver import Receiver as Receiver
from braintree.resource_collection import ResourceCollection as ResourceCollection
from braintree.risk_data import RiskData as RiskData
from braintree.samsung_pay_card import SamsungPayCard as SamsungPayCard
from braintree.search import Search as Search
from braintree.sender import Sender as Sender
from braintree.sepa_direct_debit_account import SepaDirectDebitAccount as SepaDirectDebitAccount
from braintree.settlement_batch_summary import SettlementBatchSummary as SettlementBatchSummary
from braintree.signature_service import SignatureService as SignatureService
from braintree.status_event import StatusEvent as StatusEvent
from braintree.sub_merchant import SubMerchant as SubMerchant
from braintree.subscription import Subscription as Subscription
from braintree.subscription_gateway import SubscriptionGateway as SubscriptionGateway
from braintree.subscription_search import SubscriptionSearch as SubscriptionSearch
from braintree.subscription_status_event import SubscriptionStatusEvent as SubscriptionStatusEvent
from braintree.successful_result import SuccessfulResult as SuccessfulResult
from braintree.testing_gateway import TestingGateway as TestingGateway
from braintree.three_d_secure_info import ThreeDSecureInfo as ThreeDSecureInfo
from braintree.transaction import Transaction as Transaction
from braintree.transaction_amounts import TransactionAmounts as TransactionAmounts
from braintree.transaction_details import TransactionDetails as TransactionDetails
from braintree.transaction_gateway import TransactionGateway as TransactionGateway
from braintree.transaction_line_item import TransactionLineItem as TransactionLineItem
from braintree.transaction_search import TransactionSearch as TransactionSearch
from braintree.transaction_us_bank_account_request import TransactionUsBankAccountRequest as TransactionUsBankAccountRequest
from braintree.transfer import Transfer as Transfer
from braintree.unknown_payment_method import UnknownPaymentMethod as UnknownPaymentMethod
from braintree.us_bank_account import UsBankAccount as UsBankAccount
from braintree.us_bank_account_verification import UsBankAccountVerification as UsBankAccountVerification
from braintree.validation_error_collection import ValidationErrorCollection as ValidationErrorCollection
from braintree.venmo_account import VenmoAccount as VenmoAccount
from braintree.venmo_profile_data import VenmoProfileData as VenmoProfileData
from braintree.version import Version as Version
from braintree.webhook_notification import WebhookNotification as WebhookNotification
from braintree.webhook_notification_gateway import WebhookNotificationGateway as WebhookNotificationGateway
from braintree.webhook_testing import WebhookTesting as WebhookTesting
from braintree.webhook_testing_gateway import WebhookTestingGateway as WebhookTestingGateway



import io
from _typeshed import SupportsWrite
from collections.abc import Iterable, Mapping
from typing import Any
from xml.etree.ElementTree import Element

from .fragment import FragmentFD

class DataTruncatedError(Exception): ...

class FlvReader(io.BytesIO):
    def read_bytes(self, n: int) -> bytes: ...
    def read_unsigned_long_long(self) -> int: ...
    def read_unsigned_int(self) -> int: ...
    def read_unsigned_char(self) -> str: ...
    def read_string(self) -> bytes: ...
    def read_box_info(self) -> tuple[int, bytes, bytes]: ...
    def read_asrt(self) -> dict[str, Any]: ...
    def read_afrt(self) -> dict[str, Any]: ...
    def read_abst(self) -> dict[str, Any]: ...
    def read_bootstrap_info(self) -> dict[str, Any]: ...

def read_bootstrap_info(bootstrap_bytes: bytes) -> dict[str, Any]: ...
def build_fragments_list(boot_info: Mapping[str, Any]) -> list[tuple[Any, int]]: ...
def write_unsigned_int(stream: SupportsWrite[bytes], val: int) -> None: ...
def write_unsigned_int_24(stream: SupportsWrite[bytes], val: int) -> None: ...
def write_flv_header(stream: SupportsWrite[bytes]) -> None: ...
def write_metadata_tag(stream: SupportsWrite[bytes], metadata: bytes) -> None: ...
def remove_encrypted_media(media: Iterable[Element]) -> list[Any]: ...
def get_base_url(manifest: str) -> str | None: ...

class F4mFD(FragmentFD): ...



from logging import Logger

from .connector import ServiceConnector
from .rule_cache import RuleCache
from .rule_poller import RulePoller

log: Logger

class TargetPoller:
    def __init__(self, cache: RuleCache, rule_poller: RulePoller, connector: ServiceConnector) -> None: ...
    def start(self) -> None: ...



from _typeshed import Incomplete, Unused
from datetime import date, datetime
from typing import NoReturn

from braintree.credit_card import CreditCard
from braintree.error_result import ErrorResult
from braintree.resource_collection import ResourceCollection
from braintree.successful_result import SuccessfulResult

class CreditCardGateway:
    gateway: Incomplete
    config: Incomplete
    def __init__(self, gateway) -> None: ...
    def create(self, params: dict[str, Incomplete] | None = None) -> SuccessfulResult | ErrorResult | None: ...
    def delete(self, credit_card_token: str) -> SuccessfulResult: ...
    def expired(self) -> ResourceCollection: ...
    def expiring_between(self, start_date: date | datetime, end_date: date | datetime) -> ResourceCollection: ...
    def find(self, credit_card_token: str) -> CreditCard: ...
    def forward(self, credit_card_token: Unused, receiving_merchant_id: Unused) -> NoReturn: ...
    def from_nonce(self, nonce: str) -> CreditCard: ...
    def update(
        self, credit_card_token: str, params: dict[str, Incomplete] | None = None
    ) -> SuccessfulResult | ErrorResult | None: ...



from collections.abc import Iterable, Iterator, Mapping
from types import FrameType, TracebackType
from typing import Any, Generic, Literal, TypeVar
from typing_extensions import Self

class Callpoint:
    __slots__ = ("func_name", "lineno", "module_name", "module_path", "lasti", "line")
    func_name: str
    lineno: int
    module_name: str
    module_path: str
    lasti: int
    line: str
    def __init__(
        self, module_name: str, module_path: str, func_name: str, lineno: int, lasti: int, line: str | None = None
    ) -> None: ...
    def to_dict(self) -> dict[str, Any]: ...
    @classmethod
    def from_current(cls, level: int = 1) -> Self: ...
    @classmethod
    def from_frame(cls, frame: FrameType) -> Self: ...
    @classmethod
    def from_tb(cls, tb: TracebackType) -> Self: ...
    def tb_frame_str(self) -> str: ...

_CallpointT_co = TypeVar("_CallpointT_co", bound=Callpoint, covariant=True, default=Callpoint)

class TracebackInfo(Generic[_CallpointT_co]):
    callpoint_type: type[_CallpointT_co]
    frames: list[_CallpointT_co]
    def __init__(self, frames: list[_CallpointT_co]) -> None: ...
    @classmethod
    def from_frame(cls, frame: FrameType | None = None, level: int = 1, limit: int | None = None) -> Self: ...
    @classmethod
    def from_traceback(cls, tb: TracebackType | None = None, limit: int | None = None) -> Self: ...
    @classmethod
    def from_dict(cls, d: Mapping[Literal["frames"], list[_CallpointT_co]]) -> Self: ...
    def to_dict(self) -> dict[str, list[dict[str, _CallpointT_co]]]: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[_CallpointT_co]: ...
    def get_formatted(self) -> str: ...

_TracebackInfoT_co = TypeVar("_TracebackInfoT_co", bound=TracebackInfo, covariant=True, default=TracebackInfo)

class ExceptionInfo(Generic[_TracebackInfoT_co]):
    tb_info_type: type[_TracebackInfoT_co]
    exc_type: str
    exc_msg: str
    tb_info: _TracebackInfoT_co
    def __init__(self, exc_type: str, exc_msg: str, tb_info: _TracebackInfoT_co) -> None: ...
    @classmethod
    def from_exc_info(cls, exc_type: type[BaseException], exc_value: BaseException, traceback: TracebackType) -> Self: ...
    @classmethod
    def from_current(cls) -> Self: ...
    def to_dict(self) -> dict[str, str | dict[str, list[FrameType]]]: ...
    def get_formatted(self) -> str: ...
    def get_formatted_exception_only(self) -> str: ...

class ContextualCallpoint(Callpoint):
    local_reprs: dict[Any, Any]
    pre_lines: list[str]
    post_lines: list[str]
    def __init__(self, *a, **kw) -> None: ...
    @classmethod
    def from_frame(cls, frame: FrameType) -> Self: ...
    @classmethod
    def from_tb(cls, tb: TracebackType) -> Self: ...
    def to_dict(self) -> dict[str, Any]: ...

class ContextualTracebackInfo(TracebackInfo[ContextualCallpoint]):
    callpoint_type: type[ContextualCallpoint]

class ContextualExceptionInfo(ExceptionInfo[ContextualTracebackInfo]):
    tb_info_type: type[ContextualTracebackInfo]

def print_exception(
    etype: type[BaseException] | None,
    value: BaseException | None,
    tb: TracebackType | None,
    limit: int | None = None,
    file: str | None = None,
) -> None: ...

class ParsedException:
    exc_type: str
    exc_msg: str
    frames: list[FrameType]
    def __init__(self, exc_type_name: str, exc_msg: str, frames: Iterable[Mapping[str, Any]] | None = None) -> None: ...
    @property
    def source_file(self) -> str | None: ...
    def to_dict(self) -> dict[str, str | list[FrameType]]: ...
    def to_string(self) -> str: ...
    @classmethod
    def from_string(cls, tb_str: str) -> Self: ...

ParsedTB = ParsedException

__all__ = [
    "ExceptionInfo",
    "TracebackInfo",
    "Callpoint",
    "ContextualExceptionInfo",
    "ContextualTracebackInfo",
    "ContextualCallpoint",
    "print_exception",
    "ParsedException",
]



from _typeshed import Incomplete

from reportlab.graphics.widgetbase import Widget
from reportlab.lib.attrmap import *
from reportlab.lib.validators import *

class DotBox(Widget):
    xlabels: Incomplete
    ylabels: Incomplete
    labelFontName: str
    labelFontSize: int
    labelOffset: int
    strokeWidth: float
    gridDivWidth: Incomplete
    gridColor: Incomplete
    dotDiameter: Incomplete
    dotColor: Incomplete
    dotXPosition: int
    dotYPosition: int
    x: int
    y: int
    def __init__(self) -> None: ...
    def demo(self, drawing=None): ...
    def draw(self): ...



from _typeshed import Incomplete

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["closeness_vitality"]

@_dispatchable
def closeness_vitality(
    G: Graph[_Node], node=None, weight: str | None = None, wiener_index: float | None = None
) -> float | dict[Incomplete, Incomplete]: ...



from typing import Any, NamedTuple

from .scope import Scope
from .scope_manager import ScopeManager
from .span import Span, SpanContext

class Tracer:
    def __init__(self, scope_manager: ScopeManager | None = None) -> None: ...
    @property
    def scope_manager(self) -> ScopeManager: ...
    @property
    def active_span(self) -> Span | None: ...
    def start_active_span(
        self,
        operation_name: str,
        child_of: Span | SpanContext | None = None,
        references: list[Reference] | None = None,
        tags: dict[Any, Any] | None = None,
        start_time: float | None = None,
        ignore_active_span: bool = False,
        finish_on_close: bool = True,
    ) -> Scope: ...
    def start_span(
        self,
        operation_name: str | None = None,
        child_of: Span | SpanContext | None = None,
        references: list[Reference] | None = None,
        tags: dict[Any, Any] | None = None,
        start_time: float | None = None,
        ignore_active_span: bool = False,
    ) -> Span: ...
    def inject(self, span_context: SpanContext, format: str, carrier: dict[Any, Any]) -> None: ...
    def extract(self, format: str, carrier: dict[Any, Any]) -> SpanContext: ...

class ReferenceType:
    CHILD_OF: str
    FOLLOWS_FROM: str

class Reference(NamedTuple):
    type: str
    referenced_context: SpanContext | None

def child_of(referenced_context: SpanContext | None = None) -> Reference: ...
def follows_from(referenced_context: SpanContext | None = None) -> Reference: ...
def start_child_span(
    parent_span: Span, operation_name: str, tags: dict[Any, Any] | None = None, start_time: float | None = None
) -> Span: ...



import os
from _typeshed import Incomplete, SupportsRead
from collections import OrderedDict
from typing import Literal, TypedDict, overload, type_check_only

import pandas as pd
from pandas._typing import Axes

from ..base import _BboxLike, _MaskLike
from ..geodataframe import GeoDataFrame

# Keep inline with GeoDataFrame.from_file and GeoSeries.from_file
@overload
def _read_file(
    filename: str | os.PathLike[str] | SupportsRead[Incomplete],
    bbox: _BboxLike | None = None,
    mask: _MaskLike | None = None,
    columns: Axes | None = None,
    rows: int | slice | None = None,
    engine: Literal["fiona", "pyogrio"] | None = None,
    *,
    ignore_geometry: Literal[False] = False,
    layer: int | str | None = None,
    encoding: str | None = None,
    **kwargs,  # depend on engine
) -> GeoDataFrame: ...
@overload
def _read_file(
    filename: str | os.PathLike[str] | SupportsRead[Incomplete],
    bbox: _BboxLike | None = None,
    mask: _MaskLike | None = None,
    columns: Axes | None = None,
    rows: int | slice | None = None,
    engine: Literal["fiona", "pyogrio"] | None = None,
    *,
    ignore_geometry: Literal[True],
    layer: int | str | None = None,
    encoding: str | None = None,
    **kwargs,  # depend on engine
) -> pd.DataFrame: ...
@type_check_only
class _Schema(TypedDict):
    geometry: str | list[str]
    properties: OrderedDict[str, str]

def infer_schema(df: GeoDataFrame) -> _Schema: ...
def _list_layers(filename: str | bytes | os.PathLike[str] | os.PathLike[bytes] | SupportsRead[Incomplete]) -> pd.DataFrame: ...



from _typeshed import Incomplete
from typing import Final

from reportlab.graphics.widgetbase import Widget

__version__: Final[str]

class EventCalendar(Widget):
    x: int
    y: int
    width: int
    height: int
    timeColWidth: Incomplete
    trackRowHeight: int
    data: Incomplete
    trackNames: Incomplete
    startTime: Incomplete
    endTime: Incomplete
    day: int
    def __init__(self) -> None: ...
    def computeSize(self) -> None: ...
    def computeStartAndEndTimes(self) -> None: ...
    def getAllTracks(self): ...
    def getRelevantTalks(self, talkList): ...
    def scaleTime(self, theTime): ...
    def getTalkRect(self, startTime, duration, trackId, text): ...
    def draw(self): ...

def test() -> None: ...



import sys
import xml.dom.minidom
from _typeshed import Incomplete
from abc import abstractmethod
from collections import Counter
from collections.abc import Callable, Generator, Iterable, Iterator, Mapping, Sequence
from typing import Any, ClassVar, Final, Literal, Protocol, SupportsIndex, TypeVar, overload, type_check_only
from typing_extensions import Self, TypeAlias, deprecated

from docutils.frontend import Values
from docutils.transforms import Transform, Transformer
from docutils.utils import Reporter

_N = TypeVar("_N", bound=Node)
_ElementLikeType: TypeAlias = type[Element | Text | Body | Bibliographic | Inline]
_ContentModelCategory: TypeAlias = _ElementLikeType | tuple[_ElementLikeType, ...]
_ContentModelQuantifier: TypeAlias = Literal[".", "?", "+", "*"]
_ContentModelItem: TypeAlias = tuple[_ContentModelCategory, _ContentModelQuantifier]
_ContentModelTuple: TypeAlias = tuple[_ContentModelItem, ...]

@type_check_only
class _DomModule(Protocol):
    Document: type[xml.dom.minidom.Document]

__docformat__: Final = "reStructuredText"

# Functional Node Base Classes

class Node:
    # children is initialized by the subclasses
    children: Sequence[Node]
    # TODO: `parent` is actually `Element | None``, but `None`` only happens rarely,
    #       i.e. for synthetic nodes (or `document`, where it is overridden).
    #       See https://github.com/python/typeshed/blob/main/CONTRIBUTING.md#the-any-trick
    parent: Element | Any
    source: str | None
    line: int | None
    @property
    def document(self) -> document | None: ...
    @document.setter
    def document(self, value: document) -> None: ...
    def __bool__(self) -> Literal[True]: ...
    def asdom(
        self, dom: _DomModule | None = None
    ) -> xml.dom.minidom.Document | xml.dom.minidom.Element | xml.dom.minidom.Text: ...
    # While docutils documents the Node class to be abstract it does not
    # actually use the ABCMeta metaclass. We still set @abstractmethod here
    # (although it's not used in the docutils implementation) because it
    # makes Mypy reject Node() with "Cannot instantiate abstract class".
    @abstractmethod
    def copy(self) -> Self: ...
    @abstractmethod
    def deepcopy(self) -> Self: ...
    @abstractmethod
    def pformat(self, indent: str = "    ", level: int = 0) -> str: ...
    @abstractmethod
    def astext(self) -> str: ...
    def setup_child(self, child: Node) -> None: ...
    def walk(self, visitor: NodeVisitor) -> bool: ...
    def walkabout(self, visitor: NodeVisitor) -> bool: ...
    @overload
    def findall(
        self, condition: type[_N], include_self: bool = True, descend: bool = True, siblings: bool = False, ascend: bool = False
    ) -> Generator[_N]: ...
    @overload
    def findall(
        self,
        condition: Callable[[Node], bool] | None = None,
        include_self: bool = True,
        descend: bool = True,
        siblings: bool = False,
        ascend: bool = False,
    ) -> Generator[Node]: ...
    @overload
    @deprecated("The `nodes.Node.traverse()` is deprecated. Use `Node.findall()` instead.")
    def traverse(
        self, condition: type[_N], include_self: bool = True, descend: bool = True, siblings: bool = False, ascend: bool = False
    ) -> list[_N]: ...
    @overload
    @deprecated("The `nodes.Node.traverse()` is deprecated. Use `Node.findall()` instead.")
    def traverse(
        self,
        condition: Callable[[Node], bool] | None = None,
        include_self: bool = True,
        descend: bool = True,
        siblings: bool = False,
        ascend: bool = False,
    ) -> list[Node]: ...
    @overload
    def next_node(
        self, condition: type[_N], include_self: bool = False, descend: bool = True, siblings: bool = False, ascend: bool = False
    ) -> _N | None: ...
    @overload
    def next_node(
        self,
        condition: Callable[[Node], bool] | None = None,
        include_self: bool = False,
        descend: bool = True,
        siblings: bool = False,
        ascend: bool = False,
    ) -> Node | None: ...
    def validate(self, recursive: bool = True) -> None: ...
    def validate_position(self) -> None: ...

# Left out
# - def ensure_str (deprecated)
# - def unescape (canonical import from docutils.utils)
def unescape(text: str, restore_backslashes: bool = False, respect_whitespace: bool = False) -> str: ...

class Text(Node, str):
    tagname: ClassVar[str]
    children: tuple[()]

    # we omit the rawsource parameter because it has been deprecated and is ignored
    def __new__(cls, data: str) -> Self: ...
    def shortrepr(self, maxlen: int = 18) -> str: ...
    def copy(self) -> Self: ...
    def deepcopy(self) -> Self: ...
    def pformat(self, indent: str = "    ", level: int = 0) -> str: ...
    def astext(self) -> str: ...
    def rstrip(self, chars: str | None = None) -> str: ...
    def lstrip(self, chars: str | None = None) -> str: ...

_T = TypeVar("_T")

class Element(Node):
    local_attributes: ClassVar[Sequence[str]]
    valid_attributes: ClassVar[Sequence[str]]
    common_attributes: ClassVar[Sequence[str]]
    basic_attributes: ClassVar[Sequence[str]]
    list_attributes: ClassVar[Sequence[str]]
    known_attributes: ClassVar[Sequence[str]]
    content_model: ClassVar[_ContentModelTuple]
    tagname: str
    child_text_separator: ClassVar[str]
    attributes: dict[str, Any]
    children: list[Node]
    rawsource: str
    def __init__(self, rawsource: str = "", *children: Node, **attributes: Any) -> None: ...
    def shortrepr(self) -> str: ...
    def starttag(self, quoteattr: Callable[[str], str] | None = None) -> str: ...
    def endtag(self) -> str: ...
    def emptytag(self) -> str: ...
    def __len__(self) -> int: ...
    def __contains__(self, key: str | Node) -> bool: ...
    @overload
    def __getitem__(self, key: str) -> Any: ...
    @overload
    def __getitem__(self, key: int) -> Node: ...
    @overload
    def __getitem__(self, key: slice) -> list[Node]: ...
    @overload
    def __setitem__(self, key: str, item: Any) -> None: ...
    @overload
    def __setitem__(self, key: int, item: Node) -> None: ...
    @overload
    def __setitem__(self, key: slice, item: Iterable[Node]) -> None: ...
    def __delitem__(self, key: str | int | slice) -> None: ...
    def __add__(self, other: list[Node]) -> list[Node]: ...
    def __radd__(self, other: list[Node]) -> list[Node]: ...
    def __iadd__(self, other: Node | Iterable[Node]) -> Self: ...
    def astext(self) -> str: ...
    def non_default_attributes(self) -> dict[str, Any]: ...
    def attlist(self) -> list[tuple[str, Any]]: ...
    @overload
    def get(self, key: str) -> Any: ...
    @overload
    def get(self, key: str, failobj: _T) -> _T: ...
    def hasattr(self, attr: str) -> bool: ...
    def delattr(self, attr: str) -> None: ...
    @overload
    def setdefault(self, key: str) -> Any: ...
    @overload
    def setdefault(self, key: str, failobj: _T) -> Any | _T: ...
    has_key = hasattr
    def get_language_code(self, fallback: str = "") -> str: ...
    def append(self, item: Node) -> None: ...
    def extend(self, item: Iterable[Node]) -> None: ...
    def insert(self, index: SupportsIndex, item: Node | Iterable[Node] | None) -> None: ...
    def pop(self, i: int = -1) -> Node: ...
    def remove(self, item: Node) -> None: ...
    def index(self, item: Node, start: int = 0, stop: int = sys.maxsize) -> int: ...
    def previous_sibling(self) -> Node | None: ...
    def section_hierarchy(self) -> list[section]: ...
    def is_not_default(self, key: str) -> bool: ...
    def update_basic_atts(self, dict_: Mapping[str, Any] | Node) -> None: ...
    def append_attr_list(self, attr: str, values: Iterable[Any]) -> None: ...
    def coerce_append_attr_list(self, attr: str, value) -> None: ...
    def replace_attr(self, attr: str, value: Any, force: bool = True) -> None: ...
    def copy_attr_convert(self, attr: str, value: Any, replace: bool = True) -> None: ...
    def copy_attr_coerce(self, attr: str, value: Any, replace: bool) -> None: ...
    def copy_attr_concatenate(self, attr: str, value: Any, replace: bool) -> None: ...
    def copy_attr_consistent(self, attr: str, value: Any, replace: bool) -> None: ...
    def update_all_atts(
        self,
        dict_: Mapping[str, Any] | Node,
        update_fun: Callable[[Element, str, Any, bool], object] = ...,
        replace: bool = True,
        and_source: bool = False,
    ) -> None: ...
    def update_all_atts_consistantly(
        self, dict_: Mapping[str, Any] | Node, replace: bool = True, and_source: bool = False
    ) -> None: ...
    def update_all_atts_concatenating(
        self, dict_: dict[str, Any] | Node, replace: bool = True, and_source: bool = False
    ) -> None: ...
    def update_all_atts_coercion(
        self, dict_: Mapping[str, Any] | Node, replace: bool = True, and_source: bool = False
    ) -> None: ...
    def update_all_atts_convert(self, dict_: Mapping[str, Any] | Node, and_source: bool = False) -> None: ...
    def clear(self) -> None: ...
    def replace(self, old: Node, new: Node | Sequence[Node]) -> None: ...
    def replace_self(self, new: Node | Sequence[Node]) -> None: ...
    def first_child_matching_class(
        self, childclass: type[Node] | tuple[type[Node], ...], start: int = 0, end: int = sys.maxsize
    ) -> int | None: ...
    def first_child_not_matching_class(
        self, childclass: type[Node] | tuple[type[Node], ...], start: int = 0, end: int = sys.maxsize
    ) -> int | None: ...
    def pformat(self, indent: str = "    ", level: int = 0) -> str: ...
    def copy(self) -> Self: ...
    def deepcopy(self) -> Self: ...
    def note_referenced_by(self, name: str | None = None, id: str | None = None) -> None: ...
    @classmethod
    def is_not_list_attribute(cls, attr: str) -> bool: ...
    @classmethod
    def is_not_known_attribute(cls, attr: str) -> bool: ...
    def validate_attributes(self) -> None: ...
    def validate_content(
        self, model: _ContentModelTuple | None = None, elements: Sequence[Incomplete] | None = None
    ) -> list[Incomplete]: ...

    # '__iter__' is added as workaround, since mypy doesn't support classes that are iterable via '__getitem__'
    # see https://github.com/python/typeshed/pull/10099#issuecomment-1528789395
    def __iter__(self) -> Iterator[Node]: ...

class TextElement(Element):
    def __init__(self, rawsource: str = "", text: str = "", *children: Node, **attributes) -> None: ...

class FixedTextElement(TextElement): ...
class PureTextElement(TextElement): ...

# Mixins

class Resolvable:
    resolved: int

class BackLinkable:
    list_attributes: ClassVar[Sequence[str]]
    valid_attributes: ClassVar[Sequence[str]]
    def add_backref(self, refid: str) -> None: ...

# Element Categories

class Root: ...
class Titular: ...
class PreBibliographic: ...
class Bibliographic: ...

class Decorative(PreBibliographic):
    content_model: ClassVar[_ContentModelTuple]

class Structural: ...
class SubStructural: ...
class Body: ...
class General(Body): ...
class Sequential(Body): ...

class Admonition(Body):
    content_model: ClassVar[_ContentModelTuple]

class Special(Body): ...
class Invisible(PreBibliographic): ...
class Part: ...
class Inline: ...
class Referential(Resolvable): ...

class Targetable(Resolvable):
    referenced: int
    indirect_reference_name: str | None

class Labeled: ...

# Root Element

_Document: TypeAlias = document
_Decoration: TypeAlias = decoration

class document(Root, Structural, Element):
    current_source: str | None
    current_line: int | None
    settings: Values
    reporter: Reporter
    indirect_targets: list[target]
    substitution_defs: dict[str, substitution_definition]
    substitution_names: dict[str, str]
    refnames: dict[str, list[Element]]
    refids: dict[str, list[Element]]
    nameids: dict[str, str]
    nametypes: dict[str, bool]
    ids: dict[str, Element]
    footnote_refs: dict[str, list[footnote_reference]]
    citation_refs: dict[str, list[citation_reference]]
    autofootnotes: list[footnote]
    autofootnote_refs: list[footnote_reference]
    symbol_footnotes: list[footnote]
    symbol_footnote_refs: list[footnote_reference]
    footnotes: list[footnote]
    citations: list[citation]
    autofootnote_start: int
    symbol_footnote_start: int
    id_counter: Counter[int]
    parse_messages: list[system_message]
    transform_messages: list[system_message]
    transformer: Transformer
    decoration: decoration | None
    document: Self
    def __init__(self, settings: Values, reporter: Reporter, *args: Node, **kwargs: Any) -> None: ...
    def asdom(self, dom: Any | None = None) -> Any: ...
    def set_id(self, node: Element, msgnode: Element | None = None, suggested_prefix: str = "") -> str: ...
    def set_name_id_map(self, node: Element, id: str, msgnode: Element | None = None, explicit: bool = False) -> None: ...
    def set_duplicate_name_id(self, node: Element, id: str, name: str, msgnode: Element, explicit: bool) -> None: ...
    def has_name(self, name: str) -> bool: ...
    def note_implicit_target(self, target: Element, msgnode: Element | None = None) -> None: ...
    def note_explicit_target(self, target: Element, msgnode: Element | None = None) -> None: ...
    def note_refname(self, node: Element) -> None: ...
    def note_refid(self, node: Element) -> None: ...
    def note_indirect_target(self, target: target) -> None: ...
    def note_anonymous_target(self, target: target) -> None: ...
    def note_autofootnote(self, footnote: footnote) -> None: ...
    def note_autofootnote_ref(self, ref: footnote_reference) -> None: ...
    def note_symbol_footnote(self, footnote: footnote) -> None: ...
    def note_symbol_footnote_ref(self, ref: footnote_reference) -> None: ...
    def note_footnote(self, footnote: footnote) -> None: ...
    def note_footnote_ref(self, ref: footnote_reference) -> None: ...
    def note_citation(self, citation: citation) -> None: ...
    def note_citation_ref(self, ref: citation_reference) -> None: ...
    def note_substitution_def(self, subdef: substitution_definition, def_name: str, msgnode: Element | None = None) -> None: ...
    def note_substitution_ref(self, subref: substitution_reference, refname: str) -> None: ...
    def note_pending(self, pending: pending, priority: int | None = None) -> None: ...
    def note_parse_message(self, message: system_message) -> None: ...
    def note_transform_message(self, message: system_message) -> None: ...
    def note_source(self, source: str, offset: int) -> None: ...
    def copy(self) -> Self: ...
    def get_decoration(self) -> _Decoration: ...

# Title Elements

class title(Titular, PreBibliographic, TextElement): ...
class subtitle(Titular, PreBibliographic, TextElement): ...
class rubric(Titular, TextElement): ...

# Meta-Data Element

class meta(PreBibliographic, Element): ...

# Bibliographic Elements

class docinfo(Bibliographic, Element): ...
class author(Bibliographic, TextElement): ...
class authors(Bibliographic, Element): ...
class organization(Bibliographic, TextElement): ...
class address(Bibliographic, FixedTextElement): ...
class contact(Bibliographic, TextElement): ...
class version(Bibliographic, TextElement): ...
class revision(Bibliographic, TextElement): ...
class status(Bibliographic, TextElement): ...
class date(Bibliographic, TextElement): ...
class copyright(Bibliographic, TextElement): ...

# Decorative Elements

class decoration(Decorative, Element):
    def get_header(self) -> header: ...
    def get_footer(self) -> footer: ...

class header(Decorative, Element): ...
class footer(Decorative, Element): ...

# Structural Elements

class section(Structural, Element): ...
class topic(Structural, Element): ...
class sidebar(Structural, Element): ...
class transition(Structural, Element): ...

# Body Elements
# ===============

class paragraph(General, TextElement): ...
class compound(General, Element): ...
class container(General, Element): ...
class bullet_list(Sequential, Element): ...
class enumerated_list(Sequential, Element): ...
class list_item(Part, Element): ...
class definition_list(Sequential, Element): ...
class definition_list_item(Part, Element): ...
class term(Part, TextElement): ...
class classifier(Part, TextElement): ...
class definition(Part, Element): ...
class field_list(Sequential, Element): ...
class field(Part, Element): ...
class field_name(Part, TextElement): ...
class field_body(Part, Element): ...
class option(Part, Element): ...
class option_argument(Part, TextElement): ...
class option_group(Part, Element): ...
class option_list(Sequential, Element): ...
class option_list_item(Part, Element): ...
class option_string(Part, TextElement): ...
class description(Part, Element): ...
class literal_block(General, FixedTextElement): ...
class doctest_block(General, FixedTextElement): ...
class math_block(General, FixedTextElement): ...
class line_block(General, Element): ...

class line(Part, TextElement):
    indent: str | None

class block_quote(General, Element): ...
class attribution(Part, TextElement): ...
class attention(Admonition, Element): ...
class caution(Admonition, Element): ...
class danger(Admonition, Element): ...
class error(Admonition, Element): ...
class important(Admonition, Element): ...
class note(Admonition, Element): ...
class tip(Admonition, Element): ...
class hint(Admonition, Element): ...
class warning(Admonition, Element): ...
class admonition(Admonition, Element): ...
class comment(Special, Invisible, FixedTextElement): ...
class substitution_definition(Special, Invisible, TextElement): ...
class target(Special, Invisible, Inline, TextElement, Targetable): ...
class footnote(General, BackLinkable, Element, Labeled, Targetable): ...
class citation(General, BackLinkable, Element, Labeled, Targetable): ...
class label(Part, TextElement): ...
class figure(General, Element): ...
class caption(Part, TextElement): ...
class legend(Part, Element): ...
class table(General, Element): ...
class tgroup(Part, Element): ...

class colspec(Part, Element):
    def propwidth(self) -> float: ...

class thead(Part, Element): ...
class tbody(Part, Element): ...
class row(Part, Element): ...
class entry(Part, Element): ...

class system_message(Special, BackLinkable, PreBibliographic, Element):
    def __init__(self, message: str | None = None, *children: Node, **attributes) -> None: ...
    def astext(self) -> str: ...

class pending(Special, Invisible, Element):
    transform: type[Transform]
    details: Mapping[str, Any]
    def __init__(
        self,
        transform: type[Transform],
        details: Mapping[str, Any] | None = None,
        rawsource: str = "",
        *children: Node,
        **attributes,
    ) -> None: ...

class raw(Special, Inline, PreBibliographic, FixedTextElement): ...

# Inline Elements

class emphasis(Inline, TextElement): ...
class strong(Inline, TextElement): ...
class literal(Inline, TextElement): ...
class reference(General, Inline, Referential, TextElement): ...
class footnote_reference(Inline, Referential, TextElement): ...
class citation_reference(Inline, Referential, TextElement): ...
class substitution_reference(Inline, TextElement): ...
class title_reference(Inline, TextElement): ...
class abbreviation(Inline, TextElement): ...
class acronym(Inline, TextElement): ...
class superscript(Inline, TextElement): ...
class subscript(Inline, TextElement): ...
class math(Inline, TextElement): ...
class image(General, Inline, Element): ...
class inline(Inline, TextElement): ...
class problematic(Inline, TextElement): ...
class generated(Inline, TextElement): ...

# Auxiliary Classes, Functions, and Data

node_class_names: list[str]

class NodeVisitor:
    optional: ClassVar[tuple[str, ...]]
    document: _Document
    def __init__(self, document: _Document) -> None: ...
    def dispatch_visit(self, node: Node) -> Any: ...
    def dispatch_departure(self, node: Node) -> Any: ...
    def unknown_visit(self, node: Node) -> Any: ...
    def unknown_departure(self, node: Node) -> Any: ...

    # These methods only exist on the subclasses `GenericNodeVisitor` and `SparseNodeVisitor` at runtime.
    # If subclassing `NodeVisitor` directly, `visit_*` methods must be implemented for nodes and children that will be called
    # with `Node.walk()` and `Node.walkabout()`.
    # `depart_*` methods must also be implemented for nodes and children that will be called with `Node.walkabout()`.
    def visit_Text(self, node: Text) -> Any: ...
    def visit_abbreviation(self, node: abbreviation) -> Any: ...
    def visit_acronym(self, node: acronym) -> Any: ...
    def visit_address(self, node: address) -> Any: ...
    def visit_admonition(self, node: admonition) -> Any: ...
    def visit_attention(self, node: attention) -> Any: ...
    def visit_attribution(self, node: attribution) -> Any: ...
    def visit_author(self, node: author) -> Any: ...
    def visit_authors(self, node: authors) -> Any: ...
    def visit_block_quote(self, node: block_quote) -> Any: ...
    def visit_bullet_list(self, node: bullet_list) -> Any: ...
    def visit_caption(self, node: caption) -> Any: ...
    def visit_caution(self, node: caution) -> Any: ...
    def visit_citation(self, node: citation) -> Any: ...
    def visit_citation_reference(self, node: citation_reference) -> Any: ...
    def visit_classifier(self, node: classifier) -> Any: ...
    def visit_colspec(self, node: colspec) -> Any: ...
    def visit_comment(self, node: comment) -> Any: ...
    def visit_compound(self, node: compound) -> Any: ...
    def visit_contact(self, node: contact) -> Any: ...
    def visit_container(self, node: container) -> Any: ...
    def visit_copyright(self, node: copyright) -> Any: ...
    def visit_danger(self, node: danger) -> Any: ...
    def visit_date(self, node: date) -> Any: ...
    def visit_decoration(self, node: decoration) -> Any: ...
    def visit_definition(self, node: definition) -> Any: ...
    def visit_definition_list(self, node: definition_list) -> Any: ...
    def visit_definition_list_item(self, node: definition_list_item) -> Any: ...
    def visit_description(self, node: description) -> Any: ...
    def visit_docinfo(self, node: docinfo) -> Any: ...
    def visit_doctest_block(self, node: doctest_block) -> Any: ...
    def visit_document(self, node: _Document) -> Any: ...
    def visit_emphasis(self, node: emphasis) -> Any: ...
    def visit_entry(self, node: entry) -> Any: ...
    def visit_enumerated_list(self, node: enumerated_list) -> Any: ...
    def visit_error(self, node: error) -> Any: ...
    def visit_field(self, node: field) -> Any: ...
    def visit_field_body(self, node: field_body) -> Any: ...
    def visit_field_list(self, node: field_list) -> Any: ...
    def visit_field_name(self, node: field_name) -> Any: ...
    def visit_figure(self, node: figure) -> Any: ...
    def visit_footer(self, node: footer) -> Any: ...
    def visit_footnote(self, node: footnote) -> Any: ...
    def visit_footnote_reference(self, node: footnote_reference) -> Any: ...
    def visit_generated(self, node: generated) -> Any: ...
    def visit_header(self, node: header) -> Any: ...
    def visit_hint(self, node: hint) -> Any: ...
    def visit_image(self, node: image) -> Any: ...
    def visit_important(self, node: important) -> Any: ...
    def visit_inline(self, node: inline) -> Any: ...
    def visit_label(self, node: label) -> Any: ...
    def visit_legend(self, node: legend) -> Any: ...
    def visit_line(self, node: line) -> Any: ...
    def visit_line_block(self, node: line_block) -> Any: ...
    def visit_list_item(self, node: list_item) -> Any: ...
    def visit_literal(self, node: literal) -> Any: ...
    def visit_literal_block(self, node: literal_block) -> Any: ...
    def visit_math(self, node: math) -> Any: ...
    def visit_math_block(self, node: math_block) -> Any: ...
    def visit_meta(self, node: meta) -> Any: ...
    def visit_note(self, node: note) -> Any: ...
    def visit_option(self, node: option) -> Any: ...
    def visit_option_argument(self, node: option_argument) -> Any: ...
    def visit_option_group(self, node: option_group) -> Any: ...
    def visit_option_list(self, node: option_list) -> Any: ...
    def visit_option_list_item(self, node: option_list_item) -> Any: ...
    def visit_option_string(self, node: option_string) -> Any: ...
    def visit_organization(self, node: organization) -> Any: ...
    def visit_paragraph(self, node: paragraph) -> Any: ...
    def visit_pending(self, node: pending) -> Any: ...
    def visit_problematic(self, node: problematic) -> Any: ...
    def visit_raw(self, node: raw) -> Any: ...
    def visit_reference(self, node: reference) -> Any: ...
    def visit_revision(self, node: revision) -> Any: ...
    def visit_row(self, node: row) -> Any: ...
    def visit_rubric(self, node: rubric) -> Any: ...
    def visit_section(self, node: section) -> Any: ...
    def visit_sidebar(self, node: sidebar) -> Any: ...
    def visit_status(self, node: status) -> Any: ...
    def visit_strong(self, node: strong) -> Any: ...
    def visit_subscript(self, node: subscript) -> Any: ...
    def visit_substitution_definition(self, node: substitution_definition) -> Any: ...
    def visit_substitution_reference(self, node: substitution_reference) -> Any: ...
    def visit_subtitle(self, node: subtitle) -> Any: ...
    def visit_superscript(self, node: superscript) -> Any: ...
    def visit_system_message(self, node: system_message) -> Any: ...
    def visit_table(self, node: table) -> Any: ...
    def visit_target(self, node: target) -> Any: ...
    def visit_tbody(self, node: tbody) -> Any: ...
    def visit_term(self, node: term) -> Any: ...
    def visit_tgroup(self, node: tgroup) -> Any: ...
    def visit_thead(self, node: thead) -> Any: ...
    def visit_tip(self, node: tip) -> Any: ...
    def visit_title(self, node: title) -> Any: ...
    def visit_title_reference(self, node: title_reference) -> Any: ...
    def visit_topic(self, node: topic) -> Any: ...
    def visit_transition(self, node: transition) -> Any: ...
    def visit_version(self, node: version) -> Any: ...
    def visit_warning(self, node: warning) -> Any: ...
    def depart_Text(self, node: Text) -> Any: ...
    def depart_abbreviation(self, node: abbreviation) -> Any: ...
    def depart_acronym(self, node: acronym) -> Any: ...
    def depart_address(self, node: address) -> Any: ...
    def depart_admonition(self, node: admonition) -> Any: ...
    def depart_attention(self, node: attention) -> Any: ...
    def depart_attribution(self, node: attribution) -> Any: ...
    def depart_author(self, node: author) -> Any: ...
    def depart_authors(self, node: authors) -> Any: ...
    def depart_block_quote(self, node: block_quote) -> Any: ...
    def depart_bullet_list(self, node: bullet_list) -> Any: ...
    def depart_caption(self, node: caption) -> Any: ...
    def depart_caution(self, node: caution) -> Any: ...
    def depart_citation(self, node: citation) -> Any: ...
    def depart_citation_reference(self, node: citation_reference) -> Any: ...
    def depart_classifier(self, node: classifier) -> Any: ...
    def depart_colspec(self, node: colspec) -> Any: ...
    def depart_comment(self, node: comment) -> Any: ...
    def depart_compound(self, node: compound) -> Any: ...
    def depart_contact(self, node: contact) -> Any: ...
    def depart_container(self, node: container) -> Any: ...
    def depart_copyright(self, node: copyright) -> Any: ...
    def depart_danger(self, node: danger) -> Any: ...
    def depart_date(self, node: date) -> Any: ...
    def depart_decoration(self, node: decoration) -> Any: ...
    def depart_definition(self, node: definition) -> Any: ...
    def depart_definition_list(self, node: definition_list) -> Any: ...
    def depart_definition_list_item(self, node: definition_list_item) -> Any: ...
    def depart_description(self, node: description) -> Any: ...
    def depart_docinfo(self, node: docinfo) -> Any: ...
    def depart_doctest_block(self, node: doctest_block) -> Any: ...
    def depart_document(self, node: _Document) -> Any: ...
    def depart_emphasis(self, node: emphasis) -> Any: ...
    def depart_entry(self, node: entry) -> Any: ...
    def depart_enumerated_list(self, node: enumerated_list) -> Any: ...
    def depart_error(self, node: error) -> Any: ...
    def depart_field(self, node: field) -> Any: ...
    def depart_field_body(self, node: field_body) -> Any: ...
    def depart_field_list(self, node: field_list) -> Any: ...
    def depart_field_name(self, node: field_name) -> Any: ...
    def depart_figure(self, node: figure) -> Any: ...
    def depart_footer(self, node: footer) -> Any: ...
    def depart_footnote(self, node: footnote) -> Any: ...
    def depart_footnote_reference(self, node: footnote_reference) -> Any: ...
    def depart_generated(self, node: generated) -> Any: ...
    def depart_header(self, node: header) -> Any: ...
    def depart_hint(self, node: hint) -> Any: ...
    def depart_image(self, node: image) -> Any: ...
    def depart_important(self, node: important) -> Any: ...
    def depart_inline(self, node: inline) -> Any: ...
    def depart_label(self, node: label) -> Any: ...
    def depart_legend(self, node: legend) -> Any: ...
    def depart_line(self, node: line) -> Any: ...
    def depart_line_block(self, node: line_block) -> Any: ...
    def depart_list_item(self, node: list_item) -> Any: ...
    def depart_literal(self, node: literal) -> Any: ...
    def depart_literal_block(self, node: literal_block) -> Any: ...
    def depart_math(self, node: math) -> Any: ...
    def depart_math_block(self, node: math_block) -> Any: ...
    def depart_meta(self, node: meta) -> Any: ...
    def depart_note(self, node: note) -> Any: ...
    def depart_option(self, node: option) -> Any: ...
    def depart_option_argument(self, node: option_argument) -> Any: ...
    def depart_option_group(self, node: option_group) -> Any: ...
    def depart_option_list(self, node: option_list) -> Any: ...
    def depart_option_list_item(self, node: option_list_item) -> Any: ...
    def depart_option_string(self, node: option_string) -> Any: ...
    def depart_organization(self, node: organization) -> Any: ...
    def depart_paragraph(self, node: paragraph) -> Any: ...
    def depart_pending(self, node: pending) -> Any: ...
    def depart_problematic(self, node: problematic) -> Any: ...
    def depart_raw(self, node: raw) -> Any: ...
    def depart_reference(self, node: reference) -> Any: ...
    def depart_revision(self, node: revision) -> Any: ...
    def depart_row(self, node: row) -> Any: ...
    def depart_rubric(self, node: rubric) -> Any: ...
    def depart_section(self, node: section) -> Any: ...
    def depart_sidebar(self, node: sidebar) -> Any: ...
    def depart_status(self, node: status) -> Any: ...
    def depart_strong(self, node: strong) -> Any: ...
    def depart_subscript(self, node: subscript) -> Any: ...
    def depart_substitution_definition(self, node: substitution_definition) -> Any: ...
    def depart_substitution_reference(self, node: substitution_reference) -> Any: ...
    def depart_subtitle(self, node: subtitle) -> Any: ...
    def depart_superscript(self, node: superscript) -> Any: ...
    def depart_system_message(self, node: system_message) -> Any: ...
    def depart_table(self, node: table) -> Any: ...
    def depart_target(self, node: target) -> Any: ...
    def depart_tbody(self, node: tbody) -> Any: ...
    def depart_term(self, node: term) -> Any: ...
    def depart_tgroup(self, node: tgroup) -> Any: ...
    def depart_thead(self, node: thead) -> Any: ...
    def depart_tip(self, node: tip) -> Any: ...
    def depart_title(self, node: title) -> Any: ...
    def depart_title_reference(self, node: title_reference) -> Any: ...
    def depart_topic(self, node: topic) -> Any: ...
    def depart_transition(self, node: transition) -> Any: ...
    def depart_version(self, node: version) -> Any: ...
    def depart_warning(self, node: warning) -> Any: ...

class SparseNodeVisitor(NodeVisitor): ...

class GenericNodeVisitor(NodeVisitor):
    def default_visit(self, node: Node) -> None: ...
    def default_departure(self, node: Node) -> None: ...

class TreeCopyVisitor(GenericNodeVisitor):
    parent_stack: list[Node]
    parent: list[Node]
    def get_tree_copy(self) -> Node: ...

class ValidationError(ValueError):
    def __init__(self, msg: str, problematic_element: Element | None = None) -> None: ...

class TreePruningException(Exception): ...
class SkipChildren(TreePruningException): ...
class SkipSiblings(TreePruningException): ...
class SkipNode(TreePruningException): ...
class SkipDeparture(TreePruningException): ...
class NodeFound(TreePruningException): ...
class StopTraversal(TreePruningException): ...

def make_id(string: str) -> str: ...
def dupname(node: Node, name: str) -> None: ...
def fully_normalize_name(name: str) -> str: ...
def whitespace_normalize_name(name: str) -> str: ...
def serial_escape(value: str) -> str: ...
def split_name_list(s: str) -> list[str]: ...
def pseudo_quoteattr(value: str) -> str: ...
def parse_measure(measure: str, unit_pattern: str = "[a-zA-Zµ]*|%?") -> tuple[float, str]: ...
def create_keyword_validator(*keywords: str) -> Callable[[str], str]: ...
def validate_identifier(value: str) -> str: ...
def validate_identifier_list(value: str | list[str]) -> list[str]: ...
def validate_measure(measure: str) -> str: ...
def validate_colwidth(measure: str | float) -> float: ...
def validate_NMTOKEN(value: str) -> str: ...
def validate_NMTOKENS(value: str | list[str]) -> list[str]: ...
def validate_refname_list(value: str | list[str]) -> list[str]: ...
def validate_yesorno(value: str | int | bool) -> bool: ...

ATTRIBUTE_VALIDATORS: dict[str, Callable[[str], Any]]



from collections.abc import Iterable
from re import Pattern
from typing import Any
from urllib.parse import ParseResult

from asgiref.typing import ASGIReceiveCallable, ASGISendCallable
from channels.consumer import _ChannelScope
from channels.generic.websocket import AsyncWebsocketConsumer
from channels.utils import _ChannelApplication

class OriginValidator:
    application: _ChannelApplication
    allowed_origins: Iterable[str | Pattern[str]]

    def __init__(self, application: _ChannelApplication, allowed_origins: Iterable[str | Pattern[str]]) -> None: ...
    async def __call__(self, scope: _ChannelScope, receive: ASGIReceiveCallable, send: ASGISendCallable) -> Any: ...
    def valid_origin(self, parsed_origin: ParseResult | None) -> bool: ...
    def validate_origin(self, parsed_origin: ParseResult | None) -> bool: ...
    def match_allowed_origin(self, parsed_origin: ParseResult | None, pattern: str | Pattern[str]) -> bool: ...
    def get_origin_port(self, origin: ParseResult | None) -> int | None: ...

def AllowedHostsOriginValidator(application: _ChannelApplication) -> OriginValidator: ...

class WebsocketDenier(AsyncWebsocketConsumer):
    async def connect(self) -> None: ...



import datetime
import zipimport
from _typeshed import Incomplete, SupportsItems
from collections.abc import Generator, Iterable
from os import PathLike
from types import TracebackType
from typing import AnyStr, Final, Literal, TypeVar, overload, type_check_only

from reportlab.lib.rltempfile import get_rl_tempdir as get_rl_tempdir, get_rl_tempfile as get_rl_tempfile

from .rl_safe_eval import (
    rl_extended_literal_eval as rl_extended_literal_eval,
    rl_safe_exec as rl_safe_exec,
    safer_globals as safer_globals,
)

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")

__version__: Final[str]

@type_check_only
class _UNSET_:
    @staticmethod
    def __bool__() -> Literal[False]: ...

__UNSET__: Final[_UNSET_]

isPyPy: bool

def isFunction(v: object) -> bool: ...
def isMethod(v: object, mt=...) -> bool: ...
def isModule(v: object) -> bool: ...
def isSeq(v: object, _st=...) -> bool: ...
def isNative(v: object) -> bool: ...

strTypes: tuple[type[str], type[bytes]]

def asBytes(v: str | bytes, enc: str = "utf8") -> bytes: ...
def asUnicode(v: str | bytes, enc: str = "utf8") -> str: ...
def asUnicodeEx(v: str | bytes, enc: str = "utf8") -> str: ...
def asNative(v: str | bytes, enc: str = "utf8") -> str: ...
def int2Byte(i: int) -> bytes: ...
def isStr(v: object) -> bool: ...
def isBytes(v: object) -> bool: ...
def isUnicode(v: object) -> bool: ...
def isClass(v: object) -> bool: ...
def isNonPrimitiveInstance(x: object) -> bool: ...
def instantiated(v: object) -> bool: ...
def bytestr(x: object, enc: str = "utf8") -> bytes: ...
def encode_label(args) -> str: ...
def decode_label(label: str): ...
def rawUnicode(s: str | bytes) -> str: ...
def rawBytes(s: str | bytes) -> bytes: ...

rl_exec = exec

def char2int(s: int | str | bytes) -> int: ...
def rl_reraise(t, v: BaseException, b: TracebackType | None = None) -> None: ...
def rl_add_builtins(**kwd) -> None: ...
def zipImported(ldr: zipimport.zipimporter | None = None) -> zipimport.zipimporter | None: ...

class CIDict(dict[_KT, _VT]):
    def __init__(self, *args, **kwds) -> None: ...
    def update(self, D: SupportsItems[_KT, _VT]) -> None: ...  # type:ignore[override]

def markfilename(filename, creatorcode=None, filetype=None): ...

__rl_loader__: Incomplete

def rl_glob(pattern: AnyStr, glob=...) -> list[AnyStr]: ...
def isFileSystemDistro() -> bool: ...
def isCompactDistro() -> bool: ...
def isSourceDistro() -> bool: ...
def normalize_path(p: PathLike[AnyStr]) -> PathLike[AnyStr]: ...
def recursiveImport(modulename, baseDir=None, noCWD: int = 0, debug: int = 0): ...

haveImages: Final[bool]

class ArgvDictValue:
    value: Incomplete
    func: Incomplete
    def __init__(self, value, func) -> None: ...

def getArgvDict(**kw): ...
def getHyphenater(hDict=None): ...
def open_for_read_by_name(name, mode: str = "b"): ...
def rlUrlRead(name): ...
def open_for_read(name, mode: str = "b"): ...
def open_and_read(name, mode: str = "b"): ...
def open_and_readlines(name, mode: str = "t"): ...
def rl_isfile(fn, os_path_isfile=...): ...
def rl_isdir(pn, os_path_isdir=..., os_path_normpath=...): ...
def rl_listdir(pn, os_path_isdir=..., os_path_normpath=..., os_listdir=...): ...
def rl_getmtime(pn, os_path_isfile=..., os_path_normpath=..., os_path_getmtime=..., time_mktime=...): ...
def __rl_get_module__(name, dir): ...
def rl_get_module(name, dir): ...

class ImageReader:
    fileName: Incomplete
    fp: Incomplete
    def __init__(self, fileName, ident=None) -> None: ...
    def identity(self) -> str: ...
    @classmethod
    def check_pil_image_size(cls, im) -> None: ...
    @classmethod
    def set_max_image_size(cls, max_image_size=None) -> None: ...
    def jpeg_fh(self) -> None: ...
    def getSize(self) -> tuple[int, int]: ...
    mode: Incomplete
    def getRGBData(self): ...
    def getImageData(self): ...
    def getTransparent(self): ...

class LazyImageReader(ImageReader): ...

def getImageData(imageFileName): ...

class DebugMemo:
    fn: Incomplete
    stdout: Incomplete
    store: Incomplete
    def __init__(
        self,
        fn: str = "rl_dbgmemo.dbg",
        mode: str = "w",
        getScript: int = 1,
        modules=(),
        capture_traceback: int = 1,
        stdout=None,
        **kw,
    ) -> None: ...
    def add(self, **kw) -> None: ...
    def dump(self) -> None: ...
    def dumps(self): ...
    def load(self) -> None: ...
    def loads(self, s) -> None: ...
    specials: Incomplete
    def show(self) -> None: ...
    def payload(self, name): ...
    def __setitem__(self, name, value) -> None: ...
    def __getitem__(self, name): ...

def flatten(L): ...
def find_locals(func, depth: int = 0): ...

class _FmtSelfDict:
    obj: Incomplete
    def __init__(self, obj, overrideArgs) -> None: ...
    def __getitem__(self, k): ...

class FmtSelfDict: ...

def simpleSplit(text: str | bytes, fontName: str | None, fontSize: float, maxWidth: float | None): ...
@overload
def escapeTextOnce(text: None) -> None: ...
@overload
def escapeTextOnce(text: str | bytes) -> str: ...
def fileName2FSEnc(fn): ...
def prev_this_next(items): ...
def commasplit(s: str | bytes) -> list[str]: ...
def commajoin(l: Iterable[str | bytes]) -> str: ...
def findInPaths(fn, paths, isfile: bool = True, fail: bool = False): ...
def annotateException(msg: str, enc: str = "utf8", postMsg: str = "", sep: str = " ") -> None: ...
def escapeOnce(data: str) -> str: ...

class IdentStr(str):
    def __new__(cls, value): ...

class RLString(str):
    def __new__(cls, v, **kwds): ...

def makeFileName(s): ...

class FixedOffsetTZ(datetime.tzinfo):
    def __init__(self, h, m, name) -> None: ...
    def utcoffset(self, dt): ...
    def tzname(self, dt): ...
    def dst(self, dt): ...

class TimeStamp:
    tzname: str
    t: Incomplete
    lt: Incomplete
    YMDhms: Incomplete
    dhh: Incomplete
    dmm: Incomplete
    def __init__(self, invariant=None) -> None: ...
    @property
    def datetime(self): ...
    @property
    def asctime(self): ...

def recursiveGetAttr(obj, name, g=None): ...
def recursiveSetAttr(obj, name, value) -> None: ...
def recursiveDelAttr(obj, name) -> None: ...
def yieldNoneSplits(L) -> Generator[Incomplete, None, None]: ...

class KlassStore:
    lim: int
    store: dict[str, type]
    def __init__(self, lim: int = 127) -> None: ...
    def add(self, k: str, v: type) -> None: ...
    def __contains__(self, k) -> bool: ...
    def __getitem__(self, k: str) -> type: ...
    def get(self, k, default=None): ...
